from __future__ import annotations

import json
import shutil
import subprocess
import sys
import time
from pathlib import Path

import wlbs_scan as ws


ROOT = Path(__file__).resolve().parents[1]
DEMO = ROOT / "demo"


def reset_wlbs(root: Path):
    wlbs_dir = root / ".wlbs"
    if wlbs_dir.exists():
        last_error = None
        for _ in range(5):
            try:
                shutil.rmtree(wlbs_dir)
                return
            except PermissionError as exc:
                last_error = exc
                time.sleep(0.05)
        if last_error:
            raise last_error


def prepare_demo_graph():
    reset_wlbs(DEMO)
    store = ws.WorldLineStore(DEMO)
    for i in range(3):
        store.record_failure("rbac", f"advisory failure {i + 1}")
    graph = ws.build_graph(DEMO)
    ws.compute_curvature(graph, store=store)
    return graph, store


def test_advise_output_schema():
    graph, store = prepare_demo_graph()
    advisory = ws.build_advisory(graph, store, "rbac")
    assert advisory["schema"] == "wlbs-advisory-v1"
    primary = advisory["advisory"]["primary_suggestion"]
    assert "tone" in primary
    assert "confidence" in primary


def test_advise_tone_is_suggestion():
    graph, store = prepare_demo_graph()
    advisory = ws.build_advisory(graph, store, "rbac")
    tone = advisory["advisory"]["primary_suggestion"]["tone"]
    assert tone in {"suggestion", "note"}
    assert tone != "directive"


def test_advise_confidence_range():
    graph, store = prepare_demo_graph()
    advisory = ws.build_advisory(graph, store, "rbac")
    confidence = advisory["advisory"]["primary_suggestion"]["confidence"]
    assert 0.0 <= confidence <= 1.0


def test_advise_routes_rbac_to_roles():
    graph, store = prepare_demo_graph()
    advisory = ws.build_advisory(graph, store, "rbac")
    assert "roles" in advisory["advisory"]["primary_suggestion"]["text"]


def test_advise_includes_open_questions():
    graph, store = prepare_demo_graph()
    advisory = ws.build_advisory(graph, store, "rbac")
    assert advisory["advisory"]["open_questions"] is not None
    assert len(advisory["advisory"]["open_questions"]) >= 1


def test_advise_json_parseable():
    reset_wlbs(DEMO)
    cmd = [
        sys.executable,
        str(ROOT / "wlbs_scan.py"),
        str(DEMO),
        "--record-failure",
        "rbac",
        "--detail",
        "cli setup failure",
    ]
    subprocess.run(cmd, cwd=ROOT, check=True, capture_output=True, text=True, encoding="utf-8")
    proc = subprocess.run(
        [sys.executable, str(ROOT / "wlbs_scan.py"), str(DEMO), "--advise", "rbac", "--json"],
        cwd=ROOT,
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    payload = json.loads(proc.stdout)
    assert payload["schema"] == "wlbs-advisory-v1"



"""Implementation of Rule L019."""

from typing import Any, Dict, Optional, Tuple

from sqlfluff.core.parser import RawSegment, WhitespaceSegment
from sqlfluff.core.rules.base import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.doc_decorators import (
    document_fix_compatible,
    document_configuration,
)


@document_fix_compatible
@document_configuration
class Rule_L019(BaseRule):
    """Leading/Trailing comma enforcement.

    | **Anti-pattern**
    | There is a mixture of leading and trailing commas.

    .. code-block:: sql

        SELECT
            a
            , b,
            c
        FROM foo

    | **Best practice**
    | By default sqlfluff prefers trailing commas, however it
    | is configurable for leading commas. Whichever option you chose
    | it does expect you to be consistent.

    .. code-block:: sql

        SELECT
            a,
            b,
            c
        FROM foo

        -- Alternatively, set the configuration file to 'leading'
        -- and then the following would be acceptable:

        SELECT
            a
            , b
            , c
        FROM foo


    """

    _works_on_unparsable = False
    config_keywords = ["comma_style"]

    @staticmethod
    def _last_comment_seg(raw_stack):
        """Trace the raw stack back to the most recent comment segment.

        A return value of `None` indicates no code segments preceding the current
        position.
        """
        for segment in raw_stack[::-1]:
            if segment.is_comment:
                return segment
        return None

    @staticmethod
    def _last_code_seg(raw_stack: Tuple[RawSegment, ...]) -> Optional[RawSegment]:
        """Trace the raw stack back to the most recent code segment.

        A return value of `None` indicates no code segments preceding the current
        position.
        """
        for segment in raw_stack[::-1]:
            if segment.is_code or segment.is_type("newline"):
                return segment
        return None

    @staticmethod
    def _get_following_seg(
        raw_stack: Tuple[RawSegment, ...], segment: RawSegment
    ) -> RawSegment:
        """Given a segment in raw_stack, return the segment following."""
        idx = raw_stack.index(segment)
        if idx < len(raw_stack):
            return raw_stack[idx + 1]
        raise ValueError("No following segment available")  # pragma: no cover

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Enforce comma placement.

        For leading commas we're looking for trailing commas, so
        we look for newline segments. For trailing commas we're
        looking for leading commas, so we look for the comma itself.

        We also want to handle proper whitespace removal/addition. We remove
        any trailing whitespace after the leading comma, when converting a
        leading comma to a trailing comma. We add whitespace after the leading
        comma when converting a trailing comma to a leading comma.
        """
        # Config type hints
        self.comma_style: str

        if not context.memory:
            memory: Dict[str, Any] = {
                # Trailing comma keys
                #
                # Do we have a fix in place for removing a leading
                # comma violation, and inserting a new trailing comma?
                "insert_trailing_comma": False,
                # A list of whitespace segments that come after a
                # leading comma violation, to be removed during fixing.
                "whitespace_deletions": None,
                # The leading comma violation segment to be removed during fixing
                "last_leading_comma_seg": None,
                # The newline segment where we're going to insert our new trailing
                # comma during fixing
                "anchor_for_new_trailing_comma_seg": None,
                #
                # Leading comma keys
                #
                # Do we have a fix in place for removing a trailing
                # comma violation, and inserting a new leading comma?
                "insert_leading_comma": False,
                # The trailing comma violation segment to be removed during fixing
                "last_trailing_comma_segment": None,
            }
        else:
            memory = context.memory

        if self.comma_style == "trailing":
            # A comma preceded by a new line == a leading comma
            if context.segment.is_type("comma"):
                last_seg = self._last_code_seg(context.raw_stack)
                if last_seg and last_seg.is_type("newline"):
                    # Recorded where the fix should be applied
                    memory["last_leading_comma_seg"] = context.segment
                    last_comment_seg = self._last_comment_seg(context.raw_stack)
                    inline_comment = (
                        last_comment_seg.pos_marker.line_no
                        == last_seg.pos_marker.line_no
                        if last_comment_seg
                        else False
                    )
                    # If we have a comment right before the newline, then anchor
                    # the fix at the comment instead
                    memory["anchor_for_new_trailing_comma_seg"] = (
                        last_seg if not inline_comment else last_comment_seg
                    )
                    # Trigger fix routine
                    memory["insert_trailing_comma"] = True
                    memory["whitespace_deletions"] = []
                    return LintResult(memory=memory)
            # Have we found a leading comma violation?
            if memory["insert_trailing_comma"]:
                # Search for trailing whitespace to delete after the leading
                # comma violation
                if context.segment.is_type("whitespace"):
                    memory["whitespace_deletions"] += [context.segment]
                    return LintResult(memory=memory)
                else:
                    # We've run out of whitespace to delete, time to fix
                    last_leading_comma_seg = memory["last_leading_comma_seg"]
                    # Scan backwards to find the last code segment, skipping
                    # over lines that are either entirely blank or just a
                    # comment. We want to place the comma immediately after it.
                    last_code_seg = None
                    while last_code_seg is None or last_code_seg.is_type("newline"):
                        last_code_seg = self._last_code_seg(
                            context.raw_stack[
                                : context.raw_stack.index(
                                    last_code_seg
                                    if last_code_seg
                                    else memory["last_leading_comma_seg"]
                                )
                            ]
                        )
                    return LintResult(
                        anchor=last_leading_comma_seg,
                        description="Found leading comma. Expected only trailing.",
                        fixes=[
                            LintFix.delete(last_leading_comma_seg),
                            *[
                                LintFix.delete(d)
                                for d in memory["whitespace_deletions"]
                            ],
                            LintFix.create_before(
                                anchor_segment=self._get_following_seg(
                                    context.raw_stack, last_code_seg
                                ),
                                edit_segments=[last_leading_comma_seg],
                            ),
                        ],
                    )

        elif self.comma_style == "leading":
            # A new line preceded by a comma == a trailing comma
            if context.segment.is_type("newline"):
                last_seg = self._last_code_seg(context.raw_stack)
                # no code precedes the current position: no issue
                if last_seg is None:
                    return None
                if last_seg.is_type("comma"):
                    # Trigger fix routine
                    memory["insert_leading_comma"] = True
                    # Record where the fix should be applied
                    memory["last_trailing_comma_segment"] = last_seg
                    return LintResult(memory=memory)
            # Have we found a trailing comma violation?
            if memory["insert_leading_comma"]:
                # Only insert the comma here if this isn't a comment/whitespace segment
                if context.segment.is_code:
                    last_comma_seg = memory["last_trailing_comma_segment"]
                    # Create whitespace to insert after the new leading comma
                    new_whitespace_seg = WhitespaceSegment()
                    return LintResult(
                        anchor=last_comma_seg,
                        description="Found trailing comma. Expected only leading.",
                        fixes=[
                            LintFix.delete(last_comma_seg),
                            LintFix.create_before(
                                anchor_segment=context.segment,
                                edit_segments=[last_comma_seg, new_whitespace_seg],
                            ),
                        ],
                    )
        # Otherwise, no issue
        return None



from pathlib import Path
import platform
import warnings

import pandas as pd
import os
from pkg_resources import parse_version
import pytest
from functools import wraps

import pvlib
from pvlib.location import Location

pvlib_base_version = \
    parse_version(parse_version(pvlib.__version__).base_version)


# decorator takes one argument: the base version for which it should fail
# for example @fail_on_pvlib_version('0.7') will cause a test to fail
# on pvlib versions 0.7a, 0.7b, 0.7rc1, etc.
def fail_on_pvlib_version(version):
    # second level of decorator takes the function under consideration
    def wrapper(func):
        # third level defers computation until the test is called
        # this allows the specific test to fail at test runtime,
        # rather than at decoration time (when the module is imported)
        @wraps(func)
        def inner(*args, **kwargs):
            # fail if the version is too high
            if pvlib_base_version >= parse_version(version):
                pytest.fail('the tested function is scheduled to be '
                            'removed in %s' % version)
            # otherwise return the function to be executed
            else:
                return func(*args, **kwargs)
        return inner
    return wrapper


def _check_pandas_assert_kwargs(kwargs):
    # handles the change in API related to default
    # tolerances in pandas 1.1.0.  See pvlib GH #1018
    if parse_version(pd.__version__) >= parse_version('1.1.0'):
        if kwargs.pop('check_less_precise', False):
            kwargs['atol'] = 1e-3
            kwargs['rtol'] = 1e-3
        else:
            kwargs['atol'] = 1e-5
            kwargs['rtol'] = 1e-5
    else:
        kwargs.pop('rtol', None)
        kwargs.pop('atol', None)
    return kwargs


def assert_index_equal(left, right, **kwargs):
    kwargs = _check_pandas_assert_kwargs(kwargs)
    pd.testing.assert_index_equal(left, right, **kwargs)


def assert_series_equal(left, right, **kwargs):
    kwargs = _check_pandas_assert_kwargs(kwargs)
    pd.testing.assert_series_equal(left, right, **kwargs)


def assert_frame_equal(left, right, **kwargs):
    kwargs = _check_pandas_assert_kwargs(kwargs)
    pd.testing.assert_frame_equal(left, right, **kwargs)


# commonly used directories in the tests
TEST_DIR = Path(__file__).parent
DATA_DIR = TEST_DIR.parent / 'data'


# pytest-rerunfailures variables
RERUNS = 5
RERUNS_DELAY = 2


platform_is_windows = platform.system() == 'Windows'
skip_windows = pytest.mark.skipif(platform_is_windows,
                                  reason='does not run on windows')


try:
    # Attempt to load BSRN credentials used for testing pvlib.iotools.get_bsrn
    bsrn_username = os.environ["BSRN_FTP_USERNAME"]
    bsrn_password = os.environ["BSRN_FTP_PASSWORD"]
    has_bsrn_credentials = True
except KeyError:
    has_bsrn_credentials = False

requires_bsrn_credentials = pytest.mark.skipif(
    not has_bsrn_credentials, reason='requires bsrn credentials')


try:
    import statsmodels  # noqa: F401
    has_statsmodels = True
except ImportError:
    has_statsmodels = False

requires_statsmodels = pytest.mark.skipif(
    not has_statsmodels, reason='requires statsmodels')


try:
    import ephem
    has_ephem = True
except ImportError:
    has_ephem = False

requires_ephem = pytest.mark.skipif(not has_ephem, reason='requires ephem')


def has_spa_c():
    try:
        from pvlib.spa_c_files.spa_py import spa_calc
    except ImportError:
        return False
    else:
        return True


requires_spa_c = pytest.mark.skipif(not has_spa_c(), reason="requires spa_c")


def has_numba():
    try:
        import numba
    except ImportError:
        return False
    else:
        vers = numba.__version__.split('.')
        if int(vers[0] + vers[1]) < 17:
            return False
        else:
            return True


requires_numba = pytest.mark.skipif(not has_numba(), reason="requires numba")

try:
    import siphon
    has_siphon = True
except ImportError:
    has_siphon = False

requires_siphon = pytest.mark.skipif(not has_siphon,
                                     reason='requires siphon')

try:
    import netCDF4  # noqa: F401
    has_netCDF4 = True
except ImportError:
    has_netCDF4 = False

requires_netCDF4 = pytest.mark.skipif(not has_netCDF4,
                                      reason='requires netCDF4')

try:
    import pvfactors  # noqa: F401
    has_pvfactors = True
except ImportError:
    has_pvfactors = False

requires_pvfactors = pytest.mark.skipif(not has_pvfactors,
                                        reason='requires pvfactors')


try:
    import PySAM  # noqa: F401
    has_pysam = True
except ImportError:
    has_pysam = False

requires_pysam = pytest.mark.skipif(not has_pysam, reason="requires PySAM")


try:
    import cftime  # noqa: F401

    has_recent_cftime = parse_version(cftime.__version__) > parse_version(
        "1.1.0"
    )
except ImportError:
    has_recent_cftime = False

requires_recent_cftime = pytest.mark.skipif(
    not has_recent_cftime, reason="requires cftime > 1.1.0"
)


@pytest.fixture()
def golden():
    return Location(39.742476, -105.1786, 'America/Denver', 1830.14)


@pytest.fixture()
def golden_mst():
    return Location(39.742476, -105.1786, 'MST', 1830.14)


@pytest.fixture()
def expected_solpos():
    return pd.DataFrame({'elevation': 39.872046,
                         'apparent_zenith': 50.111622,
                         'azimuth': 194.340241,
                         'apparent_elevation': 39.888378},
                        index=['2003-10-17T12:30:30Z'])


@pytest.fixture(scope="session")
def sam_data():
    data = {}
    with warnings.catch_warnings():
        # ignore messages about duplicate entries in the databases.
        warnings.simplefilter("ignore", UserWarning)
        data['sandiamod'] = pvlib.pvsystem.retrieve_sam('sandiamod')
        data['adrinverter'] = pvlib.pvsystem.retrieve_sam('adrinverter')
    return data


@pytest.fixture(scope="function")
def pvsyst_module_params():
    """
    Define some PVSyst module parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'gamma_ref': 1.05,
        'mu_gamma': 0.001,
        'I_L_ref': 6.0,
        'I_o_ref': 5e-9,
        'EgRef': 1.121,
        'R_sh_ref': 300,
        'R_sh_0': 1000,
        'R_s': 0.5,
        'R_sh_exp': 5.5,
        'cells_in_series': 60,
        'alpha_sc': 0.001,
    }
    return parameters


@pytest.fixture(scope='function')
def adr_inverter_parameters():
    """
    Define some ADR inverter parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'Name': 'Ablerex Electronics Co., Ltd.: ES 2200-US-240 (240Vac)'
                '[CEC 2011]',
        'Vac': 240.,
        'Pacmax': 2110.,
        'Pnom': 2200.,
        'Vnom': 396.,
        'Vmin': 155.,
        'Vmax': 413.,
        'Vdcmax': 500.,
        'MPPTHi': 450.,
        'MPPTLow': 150.,
        'Pnt': 0.25,
        'ADRCoefficients': [0.01385, 0.0152, 0.00794, 0.00286, -0.01872,
                            -0.01305, 0.0, 0.0, 0.0]
    }
    return parameters


@pytest.fixture(scope='function')
def cec_inverter_parameters():
    """
    Define some CEC inverter parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'Name': 'ABB: MICRO-0.25-I-OUTD-US-208 208V [CEC 2014]',
        'Vac': 208.0,
        'Paco': 250.0,
        'Pdco': 259.5220505,
        'Vdco': 40.24260317,
        'Pso': 1.771614224,
        'C0': -2.48e-5,
        'C1': -9.01e-5,
        'C2': 6.69e-4,
        'C3': -0.0189,
        'Pnt': 0.02,
        'Vdcmax': 65.0,
        'Idcmax': 10.0,
        'Mppt_low': 20.0,
        'Mppt_high': 50.0,
    }
    return parameters


@pytest.fixture(scope='function')
def cec_module_params():
    """
    Define some CEC module parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'Name': 'Example Module',
        'BIPV': 'Y',
        'Date': '4/28/2008',
        'T_NOCT': 65,
        'A_c': 0.67,
        'N_s': 18,
        'I_sc_ref': 7.5,
        'V_oc_ref': 10.4,
        'I_mp_ref': 6.6,
        'V_mp_ref': 8.4,
        'alpha_sc': 0.003,
        'beta_oc': -0.04,
        'a_ref': 0.473,
        'I_L_ref': 7.545,
        'I_o_ref': 1.94e-09,
        'R_s': 0.094,
        'R_sh_ref': 15.72,
        'Adjust': 10.6,
        'gamma_r': -0.5,
        'Version': 'MM105',
        'PTC': 48.9,
        'Technology': 'Multi-c-Si',
    }
    return parameters


@pytest.fixture(scope='function')
def cec_module_cs5p_220m():
    """
    Define Canadian Solar CS5P-220M module parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'Name': 'Canadian Solar CS5P-220M',
        'BIPV': 'N',
        'Date': '10/5/2009',
        'T_NOCT': 42.4,
        'A_c': 1.7,
        'N_s': 96,
        'I_sc_ref': 5.1,
        'V_oc_ref': 59.4,
        'I_mp_ref': 4.69,
        'V_mp_ref': 46.9,
        'alpha_sc': 0.004539,
        'beta_oc': -0.22216,
        'a_ref': 2.6373,
        'I_L_ref': 5.114,
        'I_o_ref': 8.196e-10,
        'R_s': 1.065,
        'R_sh_ref': 381.68,
        'Adjust': 8.7,
        'gamma_r': -0.476,
        'Version': 'MM106',
        'PTC': 200.1,
        'Technology': 'Mono-c-Si',
    }
    return parameters


@pytest.fixture(scope='function')
def cec_module_spr_e20_327():
    """
    Define SunPower SPR-E20-327 module parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'Name': 'SunPower SPR-E20-327',
        'BIPV': 'N',
        'Date': '1/14/2013',
        'T_NOCT': 46,
        'A_c': 1.631,
        'N_s': 96,
        'I_sc_ref': 6.46,
        'V_oc_ref': 65.1,
        'I_mp_ref': 5.98,
        'V_mp_ref': 54.7,
        'alpha_sc': 0.004522,
        'beta_oc': -0.23176,
        'a_ref': 2.6868,
        'I_L_ref': 6.468,
        'I_o_ref': 1.88e-10,
        'R_s': 0.37,
        'R_sh_ref': 298.13,
        'Adjust': -0.1862,
        'gamma_r': -0.386,
        'Version': 'NRELv1',
        'PTC': 301.4,
        'Technology': 'Mono-c-Si',
    }
    return parameters


@pytest.fixture(scope='function')
def cec_module_fs_495():
    """
    Define First Solar FS-495 module parameters for testing.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {
        'Name': 'First Solar FS-495',
        'BIPV': 'N',
        'Date': '9/18/2014',
        'T_NOCT': 44.6,
        'A_c': 0.72,
        'N_s': 216,
        'I_sc_ref': 1.55,
        'V_oc_ref': 86.5,
        'I_mp_ref': 1.4,
        'V_mp_ref': 67.9,
        'alpha_sc': 0.000924,
        'beta_oc': -0.22741,
        'a_ref': 2.9482,
        'I_L_ref': 1.563,
        'I_o_ref': 2.64e-13,
        'R_s': 6.804,
        'R_sh_ref': 806.27,
        'Adjust': -10.65,
        'gamma_r': -0.264,
        'Version': 'NRELv1',
        'PTC': 89.7,
        'Technology': 'CdTe',
    }
    return parameters


@pytest.fixture(scope='function')
def sapm_temperature_cs5p_220m():
    # SAPM temperature model parameters for Canadian_Solar_CS5P_220M
    # (glass/polymer) in open rack
    return {'a': -3.40641, 'b': -0.0842075, 'deltaT': 3}


@pytest.fixture(scope='function')
def sapm_module_params():
    """
    Define SAPM model parameters for Canadian Solar CS5P 220M module.

    The scope of the fixture is set to ``'function'`` to allow tests to modify
    parameters if required without affecting other tests.
    """
    parameters = {'Material': 'c-Si',
                  'Cells_in_Series': 96,
                  'Parallel_Strings': 1,
                  'A0': 0.928385,
                  'A1': 0.068093,
                  'A2': -0.0157738,
                  'A3': 0.0016606,
                  'A4': -6.93E-05,
                  'B0': 1,
                  'B1': -0.002438,
                  'B2': 0.0003103,
                  'B3': -0.00001246,
                  'B4': 2.11E-07,
                  'B5': -1.36E-09,
                  'C0': 1.01284,
                  'C1': -0.0128398,
                  'C2': 0.279317,
                  'C3': -7.24463,
                  'C4': 0.996446,
                  'C5': 0.003554,
                  'C6': 1.15535,
                  'C7': -0.155353,
                  'Isco': 5.09115,
                  'Impo': 4.54629,
                  'Voco': 59.2608,
                  'Vmpo': 48.3156,
                  'Aisc': 0.000397,
                  'Aimp': 0.000181,
                  'Bvoco': -0.21696,
                  'Mbvoc': 0.0,
                  'Bvmpo': -0.235488,
                  'Mbvmp': 0.0,
                  'N': 1.4032,
                  'IXO': 4.97599,
                  'IXXO': 3.18803,
                  'FD': 1}
    return parameters



"""
Test numerical precision of explicit single diode calculation using symbolic
mathematics. SymPy is a computer algebra system, that uses infinite precision
symbols instead of standard floating point and integer computer number types.
http://docs.sympy.org/latest/modules/evalf.html#accuracy-and-error-handling

This module can be executed from the command line to generate a high precision
dataset of I-V curve points to test the explicit single diode calculations
:func:`pvlib.singlediode.bishop88`::

    $ python test_numeric_precision.py

This generates a file in the pvlib data folder, which is specified by the
constant ``DATA_PATH``. When the test is run using ``pytest`` it will compare
the values calculated by :func:`pvlib.singlediode.bishop88` with the
high-precision values generated with SymPy.
"""

import logging
import numpy as np
import pandas as pd
from pvlib import pvsystem
from pvlib.singlediode import bishop88, estimate_voc
from .conftest import DATA_DIR

logging.basicConfig()
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)
TEST_DATA = 'bishop88_numerical_precision.csv'
DATA_PATH = DATA_DIR / TEST_DATA
POA = 888
TCELL = 55
# module parameters from CEC module SunPower SPR-E20-327
SPR_E20_327 = {
    'alpha_sc': 0.004522,
    'a_ref': 2.6868,
    'I_L_ref': 6.468,
    'I_o_ref': 1.88e-10,
    'R_s': 0.37,
    'R_sh_ref': 298.13,
}
# apply temp/irrad desoto corrections
ARGS = pvsystem.calcparams_desoto(
    effective_irradiance=POA, temp_cell=TCELL,
    EgRef=1.121, dEgdT=-0.0002677, **SPR_E20_327,
)
IL, I0, RS, RSH, NNSVTH = ARGS
IVCURVE_NPTS = 100

try:
    from sympy import symbols, exp as sy_exp
except ImportError as exc:
    LOGGER.exception(exc)
    symbols = NotImplemented
    sy_exp = NotImplemented


def generate_numerical_precision():  # pragma: no cover
    """
    Generate expected data with infinite numerical precision using SymPy.
    :return: dataframe of expected values
    """
    if symbols is NotImplemented:
        LOGGER.critical("SymPy is required to generate expected data.")
        raise ImportError("could not import sympy")
    il, io, rs, rsh, nnsvt, vd = symbols('il, io, rs, rsh, nnsvt, vd')
    a = sy_exp(vd / nnsvt)
    b = 1.0 / rsh
    i = il - io * (a - 1.0) - vd * b
    v = vd - i * rs
    c = io * a / nnsvt
    grad_i = - c - b  # di/dvd
    grad_v = 1.0 - grad_i * rs  # dv/dvd
    # dp/dv = d(iv)/dv = v * di/dv + i
    grad = grad_i / grad_v  # di/dv
    p = i * v
    grad_p = v * grad + i  # dp/dv
    grad2i = -c / nnsvt
    grad2v = -grad2i * rs
    grad2p = (
        grad_v * grad + v * (grad2i/grad_v - grad_i*grad2v/grad_v**2) + grad_i
    )
    # generate exact values
    data = dict(zip((il, io, rs, rsh, nnsvt), ARGS))
    vdtest = np.linspace(0, estimate_voc(IL, I0, NNSVTH), IVCURVE_NPTS)
    expected = []
    for test in vdtest:
        data[vd] = test
        test_data = {
            'i': np.float64(i.evalf(subs=data)),
            'v': np.float64(v.evalf(subs=data)),
            'p': np.float64(p.evalf(subs=data)),
            'grad_i': np.float64(grad_i.evalf(subs=data)),
            'grad_v': np.float64(grad_v.evalf(subs=data)),
            'grad': np.float64(grad.evalf(subs=data)),
            'grad_p': np.float64(grad_p.evalf(subs=data)),
            'grad2p': np.float64(grad2p.evalf(subs=data))
        }
        LOGGER.debug(test_data)
        expected.append(test_data)
    return pd.DataFrame(expected, index=vdtest)


def test_numerical_precision():
    """
    Test that there are no numerical errors due to floating point arithmetic.
    """
    expected = pd.read_csv(DATA_PATH)
    vdtest = np.linspace(0, estimate_voc(IL, I0, NNSVTH), IVCURVE_NPTS)
    results = bishop88(vdtest, *ARGS, gradients=True)
    assert np.allclose(expected['i'], results[0])
    assert np.allclose(expected['v'], results[1])
    assert np.allclose(expected['p'], results[2])
    assert np.allclose(expected['grad_i'], results[3])
    assert np.allclose(expected['grad_v'], results[4])
    assert np.allclose(expected['grad'], results[5])
    assert np.allclose(expected['grad_p'], results[6])
    assert np.allclose(expected['grad2p'], results[7])


if __name__ == '__main__':  # pragma: no cover
    expected = generate_numerical_precision()
    expected.to_csv(DATA_PATH)
    test_numerical_precision()



"""Implementation of Rule L032."""
from typing import List, Optional

from sqlfluff.core.rules.base import BaseRule, LintResult, RuleContext


class Rule_L032(BaseRule):
    """Prefer specifying join keys instead of using "USING".

    | **Anti-pattern**

    .. code-block:: sql

        SELECT
            table_a.field_1,
            table_b.field_2
        FROM
            table_a
        INNER JOIN table_b USING (id)

    | **Best practice**
    |  Specify the keys directly

    .. code-block:: sql

        SELECT
            table_a.field_1,
            table_b.field_2
        FROM
            table_a
        INNER JOIN table_b
            ON table_a.id = table_b.id

    """

    def _eval(self, context: RuleContext) -> Optional[List[LintResult]]:
        """Look for USING in a join clause."""
        if context.segment.is_type("join_clause"):
            for seg in context.segment.segments:
                if seg.is_type("keyword") and seg.name == "using":
                    return [
                        LintResult(
                            # Reference the element, not the string.
                            anchor=seg,
                            description=(
                                "Found USING statement. Expected only ON statements."
                            ),
                        )
                    ]
        return None



from __future__ import annotations

import importlib

_impl = importlib.import_module("wlbs_scan._impl")

for _name in dir(_impl):
    if _name.startswith("__") and _name != "__version__":
        continue
    globals()[_name] = getattr(_impl, _name)

__all__ = [name for name in globals() if not name.startswith("__")]



"""
Seasonal Tilt
=============

Example of a custom Mount class.
"""

# %%
# Some PV systems are built with the option to adjust the module
# tilt to follow seasonal changes in solar position.  For example,
# SAM calls this strategy "Seasonal Tilt".  This example shows how
# to use a custom Mount class to use the Seasonal Tilt strategy
# with :py:class:`~pvlib.modelchain.ModelChain`.

import pvlib
from pvlib import pvsystem, location, modelchain, iotools
from pvlib.temperature import TEMPERATURE_MODEL_PARAMETERS
import pandas as pd
import pathlib
import matplotlib.pyplot as plt
from dataclasses import dataclass


# %%
# New Mount classes should extend ``pvlib.pvsystem.AbstractMount``
# and must implement a ``get_orientation(solar_zenith, solar_azimuth)`` method:


@dataclass
class SeasonalTiltMount(pvsystem.AbstractMount):
    monthly_tilts: list  # length 12, one tilt per calendar month
    surface_azimuth: float = 180.0

    def get_orientation(self, solar_zenith, solar_azimuth):
        tilts = [self.monthly_tilts[m-1] for m in solar_zenith.index.month]
        return pd.DataFrame({
            'surface_tilt': tilts,
            'surface_azimuth': self.surface_azimuth,
        }, index=solar_zenith.index)


# %%
# First let's grab some weather data and make sure our mount produces tilts
# like we expect:

DATA_DIR = pathlib.Path(pvlib.__file__).parent / 'data'
tmy, metadata = iotools.read_tmy3(DATA_DIR / '723170TYA.CSV', coerce_year=1990)
# shift from TMY3 right-labeled index to left-labeled index:
tmy.index = tmy.index - pd.Timedelta(hours=1)
weather = pd.DataFrame({
    'ghi': tmy['GHI'], 'dhi': tmy['DHI'], 'dni': tmy['DNI'],
    'temp_air': tmy['DryBulb'], 'wind_speed': tmy['Wspd'],
})
loc = location.Location.from_tmy(metadata)
solpos = loc.get_solarposition(weather.index)
# same default monthly tilts as SAM:
tilts = [40, 40, 40, 20, 20, 20, 20, 20, 20, 40, 40, 40]
mount = SeasonalTiltMount(monthly_tilts=tilts)
orientation = mount.get_orientation(solpos.apparent_zenith, solpos.azimuth)
orientation['surface_tilt'].plot()
plt.ylabel('Surface Tilt [degrees]')
plt.show()

# %%
# With our custom tilt strategy defined, we can create the corresponding
# Array and PVSystem, and then run a ModelChain as usual:

module_parameters = {'pdc0': 1, 'gamma_pdc': -0.004, 'b': 0.05}
temp_params = TEMPERATURE_MODEL_PARAMETERS['sapm']['open_rack_glass_polymer']
array = pvsystem.Array(mount=mount, module_parameters=module_parameters,
                       temperature_model_parameters=temp_params)
system = pvsystem.PVSystem(arrays=[array], inverter_parameters={'pdc0': 1})
mc = modelchain.ModelChain(system, loc, spectral_model='no_loss')

_ = mc.run_model(weather)

# %%
# Now let's re-run the simulation assuming tilt=30 for the entire year:

array2 = pvsystem.Array(mount=pvsystem.FixedMount(30, 180),
                        module_parameters=module_parameters,
                        temperature_model_parameters=temp_params)
system2 = pvsystem.PVSystem(arrays=[array2], inverter_parameters={'pdc0': 1})
mc2 = modelchain.ModelChain(system2, loc, spectral_model='no_loss')
_ = mc2.run_model(weather)

# %%
# And finally, compare simulated monthly generation between the two tilt
# strategies:

# sphinx_gallery_thumbnail_number = 2
results = pd.DataFrame({
    'Seasonal 20/40 Production': mc.results.ac,
    'Fixed 30 Production': mc2.results.ac,
})
results.resample('m').sum().plot()
plt.ylabel('Monthly Production')
plt.show()



"""Sqlfluff is a SQL linter for humans."""
import sys
import pytest

# Expose the public API.
from sqlfluff.api import lint, fix, parse, list_rules, list_dialects  # noqa: F401

# Set the version attribute of the library
import pkg_resources
import configparser

# Get the current version
config = configparser.ConfigParser()
config.read([pkg_resources.resource_filename("sqlfluff", "config.ini")])

__version__ = config.get("sqlfluff", "version")

# Check major python version
if sys.version_info[0] < 3:
    raise Exception("Sqlfluff does not support Python 2. Please upgrade to Python 3.")
# Check minor python version
elif sys.version_info[1] < 6:
    raise Exception(
        "Sqlfluff %s only supports Python 3.6 and beyond. "
        "Use an earlier version of sqlfluff or a later version of Python" % __version__
    )

# Register helper functions to support variable introspection on failure.
pytest.register_assert_rewrite("sqlfluff.testing")



from numpy import isnan
import pandas as pd
import pytest

from pvlib.iotools import srml
from ..conftest import DATA_DIR, RERUNS, RERUNS_DELAY

srml_testfile = DATA_DIR / 'SRML-day-EUPO1801.txt'


def test_read_srml():
    srml.read_srml(srml_testfile)


@pytest.mark.remote_data
@pytest.mark.flaky(reruns=RERUNS, reruns_delay=RERUNS_DELAY)
def test_read_srml_remote():
    srml.read_srml('http://solardat.uoregon.edu/download/Archive/EUPO1801.txt')


def test_read_srml_columns_exist():
    data = srml.read_srml(srml_testfile)
    assert 'ghi_0' in data.columns
    assert 'ghi_0_flag' in data.columns
    assert 'dni_1' in data.columns
    assert 'dni_1_flag' in data.columns
    assert '7008' in data.columns
    assert '7008_flag' in data.columns


def test_read_srml_nans_exist():
    data = srml.read_srml(srml_testfile)
    assert isnan(data['dni_0'][1119])
    assert data['dni_0_flag'][1119] == 99


@pytest.mark.parametrize('url,year,month', [
    ('http://solardat.uoregon.edu/download/Archive/EUPO1801.txt',
     2018, 1),
    ('http://solardat.uoregon.edu/download/Archive/EUPO1612.txt',
     2016, 12),
])
@pytest.mark.remote_data
@pytest.mark.flaky(reruns=RERUNS, reruns_delay=RERUNS_DELAY)
def test_read_srml_dt_index(url, year, month):
    data = srml.read_srml(url)
    start = pd.Timestamp(f'{year:04d}{month:02d}01 00:00')
    start = start.tz_localize('Etc/GMT+8')
    end = pd.Timestamp(f'{year:04d}{month:02d}31 23:59')
    end = end.tz_localize('Etc/GMT+8')
    assert data.index[0] == start
    assert data.index[-1] == end
    assert (data.index[59::60].minute == 59).all()
    assert str(year) not in data.columns


@pytest.mark.parametrize('column,expected', [
    ('1001', 'ghi_1'),
    ('7324', '7324'),
    ('2001', '2001'),
    ('2017', 'dni_7')
])
def test_map_columns(column, expected):
    assert srml.map_columns(column) == expected


@pytest.mark.remote_data
@pytest.mark.flaky(reruns=RERUNS, reruns_delay=RERUNS_DELAY)
def test_read_srml_month_from_solardat():
    url = 'http://solardat.uoregon.edu/download/Archive/EUPO1801.txt'
    file_data = srml.read_srml(url)
    requested = srml.read_srml_month_from_solardat('EU', 2018, 1)
    assert file_data.equals(requested)


@pytest.mark.remote_data
@pytest.mark.flaky(reruns=RERUNS, reruns_delay=RERUNS_DELAY)
def test_15_minute_dt_index():
    data = srml.read_srml_month_from_solardat('TW', 2019, 4, 'RQ')
    start = pd.Timestamp('20190401 00:00')
    start = start.tz_localize('Etc/GMT+8')
    end = pd.Timestamp('20190430 23:45')
    end = end.tz_localize('Etc/GMT+8')
    assert data.index[0] == start
    assert data.index[-1] == end
    assert (data.index[3::4].minute == 45).all()


@pytest.mark.remote_data
@pytest.mark.flaky(reruns=RERUNS, reruns_delay=RERUNS_DELAY)
def test_hourly_dt_index():
    data = srml.read_srml_month_from_solardat('CD', 1986, 4, 'PH')
    start = pd.Timestamp('19860401 00:00')
    start = start.tz_localize('Etc/GMT+8')
    end = pd.Timestamp('19860430 23:00')
    end = end.tz_localize('Etc/GMT+8')
    assert data.index[0] == start
    assert data.index[-1] == end
    assert (data.index.minute == 0).all()



import numpy as np
import pandas as pd

from .conftest import assert_series_equal

from pvlib import snow
from pvlib.tools import sind


def test_fully_covered_nrel():
    dt = pd.date_range(start="2019-1-1 12:00:00", end="2019-1-1 18:00:00",
                       freq='1h')
    snowfall_data = pd.Series([1, 5, .6, 4, .23, -5, 19], index=dt)
    expected = pd.Series([False, True, False, True, False, False, True],
                         index=dt)
    fully_covered = snow.fully_covered_nrel(snowfall_data)
    assert_series_equal(expected, fully_covered)


def test_coverage_nrel_hourly():
    surface_tilt = 45
    slide_amount_coefficient = 0.197
    dt = pd.date_range(start="2019-1-1 10:00:00", end="2019-1-1 17:00:00",
                       freq='1h')
    poa_irradiance = pd.Series([400, 200, 100, 1234, 134, 982, 100, 100],
                               index=dt)
    temp_air = pd.Series([10, 2, 10, 1234, 34, 982, 10, 10], index=dt)
    snowfall_data = pd.Series([1, .5, .6, .4, .23, -5, .1, .1], index=dt)
    snow_coverage = snow.coverage_nrel(
        snowfall_data, poa_irradiance, temp_air, surface_tilt,
        threshold_snowfall=0.6)

    slide_amt = slide_amount_coefficient * sind(surface_tilt)
    covered = 1.0 - slide_amt * np.array([0, 1, 2, 3, 4, 5, 6, 7])
    expected = pd.Series(covered, index=dt)
    assert_series_equal(expected, snow_coverage)


def test_coverage_nrel_subhourly():
    surface_tilt = 45
    slide_amount_coefficient = 0.197
    dt = pd.date_range(start="2019-1-1 11:00:00", end="2019-1-1 14:00:00",
                       freq='15T')
    poa_irradiance = pd.Series([400, 200, 100, 1234, 134, 982, 100, 100, 100,
                                100, 100, 100, 0],
                               index=dt)
    temp_air = pd.Series([10, 2, 10, 1234, 34, 982, 10, 10, 10, 10, -10, -10,
                          10], index=dt)
    snowfall_data = pd.Series([1, .5, .6, .4, .23, -5, .1, .1, 0., 1., 0., 0.,
                               0.], index=dt)
    snow_coverage = snow.coverage_nrel(
        snowfall_data, poa_irradiance, temp_air, surface_tilt)
    slide_amt = slide_amount_coefficient * sind(surface_tilt) * 0.25
    covered = np.append(np.array([1., 1., 1., 1.]),
                        1.0 - slide_amt * np.array([1, 2, 3, 4, 5]))
    covered = np.append(covered, np.array([1., 1., 1., 1. - slide_amt]))
    expected = pd.Series(covered, index=dt)
    assert_series_equal(expected, snow_coverage)


def test_fully_covered_nrel_irregular():
    # test when frequency is not specified and can't be inferred
    dt = pd.DatetimeIndex(["2019-1-1 11:00:00", "2019-1-1 14:30:00",
                           "2019-1-1 15:07:00", "2019-1-1 14:00:00"])
    snowfall_data = pd.Series([1, .5, .6, .4], index=dt)
    snow_coverage = snow.fully_covered_nrel(snowfall_data,
                                            threshold_snowfall=0.5)
    covered = np.array([False, False, True, False])
    expected = pd.Series(covered, index=dt)
    assert_series_equal(expected, snow_coverage)


def test_coverage_nrel_initial():
    surface_tilt = 45
    slide_amount_coefficient = 0.197
    dt = pd.date_range(start="2019-1-1 10:00:00", end="2019-1-1 17:00:00",
                       freq='1h')
    poa_irradiance = pd.Series([400, 200, 100, 1234, 134, 982, 100, 100],
                               index=dt)
    temp_air = pd.Series([10, 2, 10, 1234, 34, 982, 10, 10], index=dt)
    snowfall_data = pd.Series([0, .5, .6, .4, .23, -5, .1, .1], index=dt)
    snow_coverage = snow.coverage_nrel(
        snowfall_data, poa_irradiance, temp_air, surface_tilt,
        initial_coverage=0.5, threshold_snowfall=1.)
    slide_amt = slide_amount_coefficient * sind(surface_tilt)
    covered = 0.5 - slide_amt * np.array([0, 1, 2, 3, 4, 5, 6, 7])
    covered = np.where(covered < 0, 0., covered)
    expected = pd.Series(covered, index=dt)
    assert_series_equal(expected, snow_coverage)


def test_dc_loss_nrel():
    num_strings = 8
    snow_coverage = pd.Series([1, 1, .5, .6, .2, .4, 0])
    expected = pd.Series([1, 1, .5, .625, .25, .5, 0])
    actual = snow.dc_loss_nrel(snow_coverage, num_strings)
    assert_series_equal(expected, actual)


def test__townsend_effective_snow():
    snow_total = np.array([25.4, 25.4, 12.7, 2.54, 0, 0, 0, 0, 0, 0, 12.7,
                           25.4])
    snow_events = np.array([2, 2, 1, 0, 0, 0, 0, 0, 0, 0, 2, 3])
    expected = np.array([19.05, 19.05, 12.7, 0, 0, 0, 0, 0, 0, 0, 9.525,
                         254 / 15])
    actual = snow._townsend_effective_snow(snow_total, snow_events)
    np.testing.assert_allclose(expected, actual, rtol=1e-07)


def test_loss_townsend():
    snow_total = np.array([25.4, 25.4, 12.7, 2.54, 0, 0, 0, 0, 0, 0, 12.7,
                           25.4])
    snow_events = np.array([2, 2, 1, 0, 0, 0, 0, 0, 0, 0, 2, 3])
    surface_tilt = 20
    relative_humidity = np.array([80, 80, 80, 80, 80, 80, 80, 80, 80, 80,
                                  80, 80])
    temp_air = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    poa_global = np.array([350000, 350000, 350000, 350000, 350000, 350000,
                           350000, 350000, 350000, 350000, 350000, 350000])
    angle_of_repose = 40
    slant_height = 2.54
    lower_edge_height = 0.254
    expected = np.array([0.07696253, 0.07992262, 0.06216201, 0.01715392, 0, 0,
                         0, 0, 0, 0, 0.02643821, 0.06068194])
    actual = snow.loss_townsend(snow_total, snow_events, surface_tilt,
                                relative_humidity, temp_air,
                                poa_global, slant_height,
                                lower_edge_height, angle_of_repose)
    np.testing.assert_allclose(expected, actual, rtol=1e-05)



"""
The ``modelchain`` module contains functions and classes that combine
many of the PV power modeling steps. These tools make it easy to
get started with pvlib and demonstrate standard ways to use the
library. With great power comes great responsibility: users should take
the time to read the source code for the module.
"""

from functools import partial
import itertools
import warnings
import pandas as pd
from dataclasses import dataclass, field
from typing import Union, Tuple, Optional, TypeVar

from pvlib import (atmosphere, clearsky, inverter, pvsystem, solarposition,
                   temperature, tools)
from pvlib.tracking import SingleAxisTracker
import pvlib.irradiance  # avoid name conflict with full import
from pvlib.pvsystem import _DC_MODEL_PARAMS
from pvlib._deprecation import pvlibDeprecationWarning
from pvlib.tools import _build_kwargs

from pvlib._deprecation import deprecated

# keys that are used to detect input data and assign data to appropriate
# ModelChain attribute
# for ModelChain.weather
WEATHER_KEYS = ('ghi', 'dhi', 'dni', 'wind_speed', 'temp_air',
                'precipitable_water')

# for ModelChain.total_irrad
POA_KEYS = ('poa_global', 'poa_direct', 'poa_diffuse')

# Optional keys to communicate temperature data. If provided,
# 'cell_temperature' overrides ModelChain.temperature_model and sets
# ModelChain.cell_temperature to the data. If 'module_temperature' is provdied,
# overrides ModelChain.temperature_model with
# pvlib.temperature.sapm_celL_from_module
TEMPERATURE_KEYS = ('module_temperature', 'cell_temperature')

DATA_KEYS = WEATHER_KEYS + POA_KEYS + TEMPERATURE_KEYS

# these dictionaries contain the default configuration for following
# established modeling sequences. They can be used in combination with
# basic_chain and ModelChain. They are used by the ModelChain methods
# ModelChain.with_pvwatts, ModelChain.with_sapm, etc.

# pvwatts documentation states that it uses the following reference for
# a temperature model: Fuentes, M. K. (1987). A Simplified Thermal Model
# for Flat-Plate Photovoltaic Arrays. SAND85-0330. Albuquerque, NM:
# Sandia National Laboratories. Accessed September 3, 2013:
# http://prod.sandia.gov/techlib/access-control.cgi/1985/850330.pdf
# pvlib python does not implement that model, so use the SAPM instead.
PVWATTS_CONFIG = dict(
    dc_model='pvwatts', ac_model='pvwatts', losses_model='pvwatts',
    transposition_model='perez', aoi_model='physical',
    spectral_model='no_loss', temperature_model='sapm'
)

SAPM_CONFIG = dict(
    dc_model='sapm', ac_model='sandia', losses_model='no_loss',
    aoi_model='sapm', spectral_model='sapm', temperature_model='sapm'
)


@deprecated(
    since='0.9.1',
    name='pvlib.modelchain.basic_chain',
    alternative=('pvlib.modelchain.ModelChain.with_pvwatts'
                 ' or pvlib.modelchain.ModelChain.with_sapm'),
    addendum='Note that the with_xyz methods take different model parameters.'
)
def basic_chain(times, latitude, longitude,
                surface_tilt, surface_azimuth,
                module_parameters, temperature_model_parameters,
                inverter_parameters,
                irradiance=None, weather=None,
                transposition_model='haydavies',
                solar_position_method='nrel_numpy',
                airmass_model='kastenyoung1989',
                altitude=None, pressure=None,
                **kwargs):
    """
    An experimental function that computes all of the modeling steps
    necessary for calculating power or energy for a PV system at a given
    location.

    Parameters
    ----------
    times : DatetimeIndex
        Times at which to evaluate the model.

    latitude : float.
        Positive is north of the equator.
        Use decimal degrees notation.

    longitude : float.
        Positive is east of the prime meridian.
        Use decimal degrees notation.

    surface_tilt : numeric
        Surface tilt angles in decimal degrees.
        The tilt angle is defined as degrees from horizontal
        (e.g. surface facing up = 0, surface facing horizon = 90)

    surface_azimuth : numeric
        Surface azimuth angles in decimal degrees.
        The azimuth convention is defined
        as degrees east of north
        (North=0, South=180, East=90, West=270).

    module_parameters : None, dict or Series
        Module parameters as defined by the SAPM. See pvsystem.sapm for
        details.

    temperature_model_parameters : None, dict or Series.
        Temperature model parameters as defined by the SAPM.
        See temperature.sapm_cell for details.

    inverter_parameters : None, dict or Series
        Inverter parameters as defined by the CEC. See
        :py:func:`inverter.sandia` for details.

    irradiance : None or DataFrame, default None
        If None, calculates clear sky data.
        Columns must be 'dni', 'ghi', 'dhi'.

    weather : None or DataFrame, default None
        If None, assumes air temperature is 20 C and
        wind speed is 0 m/s.
        Columns must be 'wind_speed', 'temp_air'.

    transposition_model : str, default 'haydavies'
        Passed to system.get_irradiance.

    solar_position_method : str, default 'nrel_numpy'
        Passed to solarposition.get_solarposition.

    airmass_model : str, default 'kastenyoung1989'
        Passed to atmosphere.relativeairmass.

    altitude : None or float, default None
        If None, computed from pressure. Assumed to be 0 m
        if pressure is also None.

    pressure : None or float, default None
        If None, computed from altitude. Assumed to be 101325 Pa
        if altitude is also None.

    **kwargs
        Arbitrary keyword arguments.
        See code for details.

    Returns
    -------
    output : (dc, ac)
        Tuple of DC power (with SAPM parameters) (DataFrame) and AC
        power (Series).
    """

    if altitude is None and pressure is None:
        altitude = 0.
        pressure = 101325.
    elif altitude is None:
        altitude = atmosphere.pres2alt(pressure)
    elif pressure is None:
        pressure = atmosphere.alt2pres(altitude)

    solar_position = solarposition.get_solarposition(
        times, latitude, longitude, altitude=altitude, pressure=pressure,
        method=solar_position_method, **kwargs)

    # possible error with using apparent zenith with some models
    airmass = atmosphere.get_relative_airmass(
        solar_position['apparent_zenith'], model=airmass_model)
    airmass = atmosphere.get_absolute_airmass(airmass, pressure)
    dni_extra = pvlib.irradiance.get_extra_radiation(solar_position.index)

    aoi = pvlib.irradiance.aoi(surface_tilt, surface_azimuth,
                               solar_position['apparent_zenith'],
                               solar_position['azimuth'])

    if irradiance is None:
        linke_turbidity = clearsky.lookup_linke_turbidity(
            solar_position.index, latitude, longitude)
        irradiance = clearsky.ineichen(
            solar_position['apparent_zenith'],
            airmass,
            linke_turbidity,
            altitude=altitude,
            dni_extra=dni_extra
        )

    total_irrad = pvlib.irradiance.get_total_irradiance(
        surface_tilt,
        surface_azimuth,
        solar_position['apparent_zenith'],
        solar_position['azimuth'],
        irradiance['dni'],
        irradiance['ghi'],
        irradiance['dhi'],
        model=transposition_model,
        dni_extra=dni_extra)

    if weather is None:
        weather = {'wind_speed': 0, 'temp_air': 20}

    cell_temperature = temperature.sapm_cell(
        total_irrad['poa_global'], weather['temp_air'], weather['wind_speed'],
        temperature_model_parameters['a'], temperature_model_parameters['b'],
        temperature_model_parameters['deltaT'])

    effective_irradiance = pvsystem.sapm_effective_irradiance(
        total_irrad['poa_direct'], total_irrad['poa_diffuse'], airmass, aoi,
        module_parameters)

    dc = pvsystem.sapm(effective_irradiance, cell_temperature,
                       module_parameters)

    ac = inverter.sandia(dc['v_mp'], dc['p_mp'], inverter_parameters)

    return dc, ac


def get_orientation(strategy, **kwargs):
    """
    Determine a PV system's surface tilt and surface azimuth
    using a named strategy.

    Parameters
    ----------
    strategy: str
        The orientation strategy.
        Allowed strategies include 'flat', 'south_at_latitude_tilt'.
    **kwargs:
        Strategy-dependent keyword arguments. See code for details.

    Returns
    -------
    surface_tilt, surface_azimuth
    """

    if strategy == 'south_at_latitude_tilt':
        surface_azimuth = 180
        surface_tilt = kwargs['latitude']
    elif strategy == 'flat':
        surface_azimuth = 180
        surface_tilt = 0
    else:
        raise ValueError('invalid orientation strategy. strategy must '
                         'be one of south_at_latitude, flat,')

    return surface_tilt, surface_azimuth


# Type for fields that vary between arrays
T = TypeVar('T')


PerArray = Union[T, Tuple[T, ...]]


@dataclass
class ModelChainResult:
    # these attributes are used in __setattr__ to determine the correct type.
    _singleton_tuples: bool = field(default=False)
    _per_array_fields = {'total_irrad', 'aoi', 'aoi_modifier',
                         'spectral_modifier', 'cell_temperature',
                         'effective_irradiance', 'dc', 'diode_params',
                         'dc_ohmic_losses', 'weather', 'albedo'}

    # system-level information
    solar_position: Optional[pd.DataFrame] = field(default=None)
    """Solar position in a DataFrame containing columns ``'apparent_zenith'``,
    ``'zenith'``, ``'apparent_elevation'``, ``'elevation'``, ``'azimuth'``
    (all in degrees), with possibly other columns depending on the solar
    position method; see :py:func:`~pvlib.solarposition.get_solarposition`
    for details."""

    airmass: Optional[pd.DataFrame] = field(default=None)
    """Air mass in a DataFrame containing columns ``'airmass_relative'``,
    ``'airmass_absolute'`` (unitless); see
    :py:meth:`~pvlib.location.Location.get_airmass` for details."""

    ac: Optional[pd.Series] = field(default=None)
    """AC power from the PV system, in a Series [W]"""

    tracking: Optional[pd.DataFrame] = field(default=None)
    """Orientation of modules on a single axis tracker, in a DataFrame with
    columns ``'surface_tilt'``, ``'surface_azimuth'``, ``'aoi'``; see
    :py:func:`~pvlib.tracking.singleaxis` for details.
    """

    losses: Optional[Union[pd.Series, float]] = field(default=None)
    """Series containing DC loss as a fraction of total DC power, as
    calculated by ``ModelChain.losses_model``.
    """

    # per DC array information
    total_irrad: Optional[PerArray[pd.DataFrame]] = field(default=None)
    """ DataFrame (or tuple of DataFrame, one for each array) containing
    columns ``'poa_global'``, ``'poa_direct'`` ``'poa_diffuse'``,
    ``poa_sky_diffuse'``, ``'poa_ground_diffuse'`` (W/m2); see
    :py:func:`~pvlib.irradiance.get_total_irradiance` for details.
    """

    aoi: Optional[PerArray[pd.Series]] = field(default=None)
    """
    Series (or tuple of Series, one for each array) containing angle of
    incidence (degrees); see :py:func:`~pvlib.irradiance.aoi` for details.
    """

    aoi_modifier: Optional[PerArray[Union[pd.Series, float]]] = \
        field(default=None)
    """Series (or tuple of Series, one for each array) containing angle of
    incidence modifier (unitless) calculated by ``ModelChain.aoi_model``,
    which reduces direct irradiance for reflections;
    see :py:meth:`~pvlib.pvsystem.PVSystem.get_iam` for details.
    """

    spectral_modifier: Optional[PerArray[Union[pd.Series, float]]] = \
        field(default=None)
    """Series (or tuple of Series, one for each array) containing spectral
    modifier (unitless) calculated by ``ModelChain.spectral_model``, which
    adjusts broadband plane-of-array irradiance for spectral content.
    """

    cell_temperature: Optional[PerArray[pd.Series]] = field(default=None)
    """Series (or tuple of Series, one for each array) containing cell
    temperature (C).
    """

    effective_irradiance: Optional[PerArray[pd.Series]] = field(default=None)
    """Series (or tuple of Series, one for each array) containing effective
    irradiance (W/m2) which is total plane-of-array irradiance adjusted for
    reflections and spectral content.
    """

    dc: Optional[PerArray[Union[pd.Series, pd.DataFrame]]] = \
        field(default=None)
    """Series or DataFrame (or tuple of Series or DataFrame, one for
    each array) containing DC power (W) for each array, calculated by
    ``ModelChain.dc_model``.
    """

    diode_params: Optional[PerArray[pd.DataFrame]] = field(default=None)
    """DataFrame (or tuple of DataFrame, one for each array) containing diode
    equation parameters (columns ``'I_L'``, ``'I_o'``, ``'R_s'``, ``'R_sh'``,
    ``'nNsVth'``, present when ModelChain.dc_model is a single diode model;
    see :py:func:`~pvlib.pvsystem.singlediode` for details.
    """

    dc_ohmic_losses: Optional[PerArray[pd.Series]] = field(default=None)
    """Series (or tuple of Series, one for each array) containing DC ohmic
    loss (W) calculated by ``ModelChain.dc_ohmic_model``.
    """

    # copies of input data, for user convenience
    weather: Optional[PerArray[pd.DataFrame]] = None
    """DataFrame (or tuple of DataFrame, one for each array) contains a
    copy of the input weather data.
    """

    times: Optional[pd.DatetimeIndex] = None
    """DatetimeIndex containing a copy of the index of the input weather data.
    """

    albedo: Optional[PerArray[pd.Series]] = None
    """Series (or tuple of Series, one for each array) containing albedo.
    """

    def _result_type(self, value):
        """Coerce `value` to the correct type according to
        ``self._singleton_tuples``."""
        # Allow None to pass through without being wrapped in a tuple
        if (self._singleton_tuples
                and not isinstance(value, tuple)
                and value is not None):
            return (value,)
        return value

    def __setattr__(self, key, value):
        if key in ModelChainResult._per_array_fields:
            value = self._result_type(value)
        super().__setattr__(key, value)


class ModelChain:
    """
    The ModelChain class to provides a standardized, high-level
    interface for all of the modeling steps necessary for calculating PV
    power from a time series of weather inputs. The same models are applied
    to all ``pvsystem.Array`` objects, so each Array must contain the
    appropriate model parameters. For example, if ``dc_model='pvwatts'``,
    then each ``Array.module_parameters`` must contain ``'pdc0'``.

    See :ref:`modelchaindoc` for examples.

    Parameters
    ----------
    system : PVSystem
        A :py:class:`~pvlib.pvsystem.PVSystem` object that represents
        the connected set of modules, inverters, etc.

    location : Location
        A :py:class:`~pvlib.location.Location` object that represents
        the physical location at which to evaluate the model.

    clearsky_model : str, default 'ineichen'
        Passed to location.get_clearsky.

    transposition_model : str, default 'haydavies'
        Passed to system.get_irradiance.

    solar_position_method : str, default 'nrel_numpy'
        Passed to location.get_solarposition.

    airmass_model : str, default 'kastenyoung1989'
        Passed to location.get_airmass.

    dc_model: None, str, or function, default None
        If None, the model will be inferred from the parameters that
        are common to all of system.arrays[i].module_parameters.
        Valid strings are 'sapm', 'desoto', 'cec', 'pvsyst', 'pvwatts'.
        The ModelChain instance will be passed as the first argument
        to a user-defined function.

    ac_model: None, str, or function, default None
        If None, the model will be inferred from the parameters that
        are common to all of system.inverter_parameters.
        Valid strings are 'sandia', 'adr', 'pvwatts'. The
        ModelChain instance will be passed as the first argument to a
        user-defined function.

    aoi_model: None, str, or function, default None
        If None, the model will be inferred from the parameters that
        are common to all of system.arrays[i].module_parameters.
        Valid strings are 'physical', 'ashrae', 'sapm', 'martin_ruiz',
        'no_loss'. The ModelChain instance will be passed as the
        first argument to a user-defined function.

    spectral_model: None, str, or function, default None
        If None, the model will be inferred from the parameters that
        are common to all of system.arrays[i].module_parameters.
        Valid strings are 'sapm', 'first_solar', 'no_loss'.
        The ModelChain instance will be passed as the first argument to
        a user-defined function.

    temperature_model: None, str or function, default None
        Valid strings are: 'sapm', 'pvsyst', 'faiman', 'fuentes', 'noct_sam'.
        The ModelChain instance will be passed as the first argument to a
        user-defined function.

    dc_ohmic_model: str or function, default 'no_loss'
        Valid strings are 'dc_ohms_from_percent', 'no_loss'. The ModelChain
        instance will be passed as the first argument to a user-defined
        function.

    losses_model: str or function, default 'no_loss'
        Valid strings are 'pvwatts', 'no_loss'. The ModelChain instance
        will be passed as the first argument to a user-defined function.

    name: None or str, default None
        Name of ModelChain instance.
    """

    # list of deprecated attributes
    _deprecated_attrs = ['solar_position', 'airmass', 'total_irrad',
                         'aoi', 'aoi_modifier', 'spectral_modifier',
                         'cell_temperature', 'effective_irradiance',
                         'dc', 'ac', 'diode_params', 'tracking',
                         'weather', 'times', 'losses']

    def __init__(self, system, location,
                 clearsky_model='ineichen',
                 transposition_model='haydavies',
                 solar_position_method='nrel_numpy',
                 airmass_model='kastenyoung1989',
                 dc_model=None, ac_model=None, aoi_model=None,
                 spectral_model=None, temperature_model=None,
                 dc_ohmic_model='no_loss',
                 losses_model='no_loss', name=None):

        self.name = name
        self.system = system

        self.location = location
        self.clearsky_model = clearsky_model
        self.transposition_model = transposition_model
        self.solar_position_method = solar_position_method
        self.airmass_model = airmass_model

        # calls setters
        self.dc_model = dc_model
        self.ac_model = ac_model
        self.aoi_model = aoi_model
        self.spectral_model = spectral_model
        self.temperature_model = temperature_model

        self.dc_ohmic_model = dc_ohmic_model
        self.losses_model = losses_model

        self.results = ModelChainResult()

    def __getattr__(self, key):
        if key in ModelChain._deprecated_attrs:
            msg = f'ModelChain.{key} is deprecated and will' \
                  f' be removed in v0.10. Use' \
                  f' ModelChain.results.{key} instead'
            warnings.warn(msg, pvlibDeprecationWarning)
            return getattr(self.results, key)
        # __getattr__ is only called if __getattribute__ fails.
        # In that case we should check if key is a deprecated attribute,
        # and fail with an AttributeError if it is not.
        raise AttributeError

    def __setattr__(self, key, value):
        if key in ModelChain._deprecated_attrs:
            msg = f'ModelChain.{key} is deprecated from v0.9. Use' \
                  f' ModelChain.results.{key} instead'
            warnings.warn(msg, pvlibDeprecationWarning)
            setattr(self.results, key, value)
        else:
            super().__setattr__(key, value)

    @classmethod
    def with_pvwatts(cls, system, location,
                     clearsky_model='ineichen',
                     airmass_model='kastenyoung1989',
                     name=None,
                     **kwargs):
        """
        ModelChain that follows the PVWatts methods.

        Parameters
        ----------
        system : PVSystem
            A :py:class:`~pvlib.pvsystem.PVSystem` object that represents
            the connected set of modules, inverters, etc.

        location : Location
            A :py:class:`~pvlib.location.Location` object that represents
            the physical location at which to evaluate the model.

        clearsky_model : str, default 'ineichen'
            Passed to location.get_clearsky.

        airmass_model : str, default 'kastenyoung1989'
            Passed to location.get_airmass.

        name: None or str, default None
            Name of ModelChain instance.

        **kwargs
            Parameters supplied here are passed to the ModelChain
            constructor and take precedence over the default
            configuration.

        Examples
        --------
        >>> module_parameters = dict(gamma_pdc=-0.003, pdc0=4500)
        >>> inverter_parameters = dict(pac0=4000)
        >>> tparams = TEMPERATURE_MODEL_PARAMETERS['sapm']['open_rack_glass_glass']
        >>> system = PVSystem(surface_tilt=30, surface_azimuth=180,
        ...     module_parameters=module_parameters,
        ...     inverter_parameters=inverter_parameters,
        ...     temperature_model_parameters=tparams)
        >>> location = Location(32.2, -110.9)
        >>> ModelChain.with_pvwatts(system, location)
        ModelChain:
          name: None
          clearsky_model: ineichen
          transposition_model: perez
          solar_position_method: nrel_numpy
          airmass_model: kastenyoung1989
          dc_model: pvwatts_dc
          ac_model: pvwatts_inverter
          aoi_model: physical_aoi_loss
          spectral_model: no_spectral_loss
          temperature_model: sapm_temp
          losses_model: pvwatts_losses
        """  # noqa: E501
        config = PVWATTS_CONFIG.copy()
        config.update(kwargs)
        return ModelChain(
            system, location,
            clearsky_model=clearsky_model,
            airmass_model=airmass_model,
            name=name,
            **config
        )

    @classmethod
    def with_sapm(cls, system, location,
                  clearsky_model='ineichen',
                  transposition_model='haydavies',
                  solar_position_method='nrel_numpy',
                  airmass_model='kastenyoung1989',
                  name=None,
                  **kwargs):
        """
        ModelChain that follows the Sandia Array Performance Model
        (SAPM) methods.

        Parameters
        ----------
        system : PVSystem
            A :py:class:`~pvlib.pvsystem.PVSystem` object that represents
            the connected set of modules, inverters, etc.

        location : Location
            A :py:class:`~pvlib.location.Location` object that represents
            the physical location at which to evaluate the model.

        clearsky_model : str, default 'ineichen'
            Passed to location.get_clearsky.

        transposition_model : str, default 'haydavies'
            Passed to system.get_irradiance.

        solar_position_method : str, default 'nrel_numpy'
            Passed to location.get_solarposition.

        airmass_model : str, default 'kastenyoung1989'
            Passed to location.get_airmass.

        name: None or str, default None
            Name of ModelChain instance.

        **kwargs
            Parameters supplied here are passed to the ModelChain
            constructor and take precedence over the default
            configuration.

        Examples
        --------
        >>> mods = pvlib.pvsystem.retrieve_sam('sandiamod')
        >>> invs = pvlib.pvsystem.retrieve_sam('cecinverter')
        >>> module_parameters = mods['Canadian_Solar_CS5P_220M___2009_']
        >>> inverter_parameters = invs['ABB__MICRO_0_25_I_OUTD_US_240__240V_']
        >>> tparams = TEMPERATURE_MODEL_PARAMETERS['sapm']['open_rack_glass_glass']
        >>> system = PVSystem(surface_tilt=30, surface_azimuth=180,
        ...     module_parameters=module_parameters,
        ...     inverter_parameters=inverter_parameters,
        ...     temperature_model_parameters=tparams)
        >>> location = Location(32.2, -110.9)
        >>> ModelChain.with_sapm(system, location)
        ModelChain:
          name: None
          clearsky_model: ineichen
          transposition_model: haydavies
          solar_position_method: nrel_numpy
          airmass_model: kastenyoung1989
          dc_model: sapm
          ac_model: snlinverter
          aoi_model: sapm_aoi_loss
          spectral_model: sapm_spectral_loss
          temperature_model: sapm_temp
          losses_model: no_extra_losses
        """  # noqa: E501
        config = SAPM_CONFIG.copy()
        config.update(kwargs)
        return ModelChain(
            system, location,
            clearsky_model=clearsky_model,
            transposition_model=transposition_model,
            solar_position_method=solar_position_method,
            airmass_model=airmass_model,
            name=name,
            **config
        )

    def __repr__(self):
        attrs = [
            'name', 'clearsky_model',
            'transposition_model', 'solar_position_method',
            'airmass_model', 'dc_model', 'ac_model', 'aoi_model',
            'spectral_model', 'temperature_model', 'losses_model'
        ]

        def getmcattr(self, attr):
            """needed to avoid recursion in property lookups"""
            out = getattr(self, attr)
            try:
                out = out.__name__
            except AttributeError:
                pass
            return out

        return ('ModelChain: \n  ' + '\n  '.join(
            f'{attr}: {getmcattr(self, attr)}' for attr in attrs))

    @property
    def dc_model(self):
        return self._dc_model

    @dc_model.setter
    def dc_model(self, model):
        # guess at model if None
        if model is None:
            self._dc_model, model = self.infer_dc_model()

        # Set model and validate parameters
        if isinstance(model, str):
            model = model.lower()
            if model in _DC_MODEL_PARAMS.keys():
                # validate module parameters
                module_parameters = tuple(
                    array.module_parameters for array in self.system.arrays)
                missing_params = (
                    _DC_MODEL_PARAMS[model] - _common_keys(module_parameters))
                if missing_params:  # some parameters are not in module.keys()
                    raise ValueError(model + ' selected for the DC model but '
                                     'one or more Arrays are missing '
                                     'one or more required parameters '
                                     ' : ' + str(missing_params))
                if model == 'sapm':
                    self._dc_model = self.sapm
                elif model == 'desoto':
                    self._dc_model = self.desoto
                elif model == 'cec':
                    self._dc_model = self.cec
                elif model == 'pvsyst':
                    self._dc_model = self.pvsyst
                elif model == 'pvwatts':
                    self._dc_model = self.pvwatts_dc
            else:
                raise ValueError(model + ' is not a valid DC power model')
        else:
            self._dc_model = partial(model, self)

    def infer_dc_model(self):
        """Infer DC power model from Array module parameters."""
        params = _common_keys(
            tuple(array.module_parameters for array in self.system.arrays))
        if {'A0', 'A1', 'C7'} <= params:
            return self.sapm, 'sapm'
        elif {'a_ref', 'I_L_ref', 'I_o_ref', 'R_sh_ref', 'R_s',
              'Adjust'} <= params:
            return self.cec, 'cec'
        elif {'a_ref', 'I_L_ref', 'I_o_ref', 'R_sh_ref', 'R_s'} <= params:
            return self.desoto, 'desoto'
        elif {'gamma_ref', 'mu_gamma', 'I_L_ref', 'I_o_ref', 'R_sh_ref',
              'R_sh_0', 'R_sh_exp', 'R_s'} <= params:
            return self.pvsyst, 'pvsyst'
        elif {'pdc0', 'gamma_pdc'} <= params:
            return self.pvwatts_dc, 'pvwatts'
        else:
            raise ValueError(
                'Could not infer DC model from the module_parameters '
                'attributes of system.arrays. Check the module_parameters '
                'attributes or explicitly set the model with the dc_model '
                'keyword argument.')

    def sapm(self):
        dc = self.system.sapm(self.results.effective_irradiance,
                              self.results.cell_temperature)
        self.results.dc = self.system.scale_voltage_current_power(dc)
        return self

    def _singlediode(self, calcparams_model_function):
        def _make_diode_params(photocurrent, saturation_current,
                               resistance_series, resistance_shunt,
                               nNsVth):
            return pd.DataFrame(
                {'I_L': photocurrent, 'I_o': saturation_current,
                 'R_s': resistance_series, 'R_sh': resistance_shunt,
                 'nNsVth': nNsVth}
            )
        params = calcparams_model_function(self.results.effective_irradiance,
                                           self.results.cell_temperature,
                                           unwrap=False)
        self.results.diode_params = tuple(itertools.starmap(
            _make_diode_params, params))
        self.results.dc = tuple(itertools.starmap(
            self.system.singlediode, params))
        self.results.dc = self.system.scale_voltage_current_power(
            self.results.dc,
            unwrap=False
        )
        self.results.dc = tuple(dc.fillna(0) for dc in self.results.dc)
        # If the system has one Array, unwrap the single return value
        # to preserve the original behavior of ModelChain
        if self.system.num_arrays == 1:
            self.results.diode_params = self.results.diode_params[0]
            self.results.dc = self.results.dc[0]
        return self

    def desoto(self):
        return self._singlediode(self.system.calcparams_desoto)

    def cec(self):
        return self._singlediode(self.system.calcparams_cec)

    def pvsyst(self):
        return self._singlediode(self.system.calcparams_pvsyst)

    def pvwatts_dc(self):
        """Calculate DC power using the PVWatts model.

        Results are stored in ModelChain.results.dc. DC power is computed
        from PVSystem.arrays[i].module_parameters['pdc0'] and then scaled by
        PVSystem.modules_per_string and PVSystem.strings_per_inverter.

        Returns
        -------
        self

        See also
        --------
        pvlib.pvsystem.PVSystem.pvwatts_dc
        pvlib.pvsystem.PVSystem.scale_voltage_current_power
        """
        dc = self.system.pvwatts_dc(
            self.results.effective_irradiance,
            self.results.cell_temperature,
            unwrap=False
        )
        p_mp = tuple(pd.DataFrame(s, columns=['p_mp']) for s in dc)
        scaled = self.system.scale_voltage_current_power(p_mp)
        self.results.dc = _tuple_from_dfs(scaled, "p_mp")
        return self

    @property
    def ac_model(self):
        return self._ac_model

    @ac_model.setter
    def ac_model(self, model):
        if model is None:
            self._ac_model = self.infer_ac_model()
        elif isinstance(model, str):
            model = model.lower()
            if model == 'sandia':
                self._ac_model = self.sandia_inverter
            elif model in 'adr':
                self._ac_model = self.adr_inverter
            elif model == 'pvwatts':
                self._ac_model = self.pvwatts_inverter
            else:
                raise ValueError(model + ' is not a valid AC power model')
        else:
            self._ac_model = partial(model, self)

    def infer_ac_model(self):
        """Infer AC power model from system attributes."""
        inverter_params = set(self.system.inverter_parameters.keys())
        if _snl_params(inverter_params):
            return self.sandia_inverter
        if _adr_params(inverter_params):
            if self.system.num_arrays > 1:
                raise ValueError(
                    'The adr inverter function cannot be used for an inverter',
                    ' with multiple MPPT inputs')
            else:
                return self.adr_inverter
        if _pvwatts_params(inverter_params):
            return self.pvwatts_inverter
        raise ValueError('could not infer AC model from '
                         'system.inverter_parameters. Check '
                         'system.inverter_parameters or explicitly '
                         'set the model with the ac_model kwarg.')

    def sandia_inverter(self):
        self.results.ac = self.system.get_ac(
            'sandia',
            _tuple_from_dfs(self.results.dc, 'p_mp'),
            v_dc=_tuple_from_dfs(self.results.dc, 'v_mp')
        )
        return self

    def adr_inverter(self):
        self.results.ac = self.system.get_ac(
            'adr',
            self.results.dc['p_mp'],
            v_dc=self.results.dc['v_mp']
        )
        return self

    def pvwatts_inverter(self):
        ac = self.system.get_ac('pvwatts', self.results.dc)
        self.results.ac = ac.fillna(0)
        return self

    @property
    def aoi_model(self):
        return self._aoi_model

    @aoi_model.setter
    def aoi_model(self, model):
        if model is None:
            self._aoi_model = self.infer_aoi_model()
        elif isinstance(model, str):
            model = model.lower()
            if model == 'ashrae':
                self._aoi_model = self.ashrae_aoi_loss
            elif model == 'physical':
                self._aoi_model = self.physical_aoi_loss
            elif model == 'sapm':
                self._aoi_model = self.sapm_aoi_loss
            elif model == 'martin_ruiz':
                self._aoi_model = self.martin_ruiz_aoi_loss
            elif model == 'no_loss':
                self._aoi_model = self.no_aoi_loss
            else:
                raise ValueError(model + ' is not a valid aoi loss model')
        else:
            self._aoi_model = partial(model, self)

    def infer_aoi_model(self):
        module_parameters = tuple(
            array.module_parameters for array in self.system.arrays)
        params = _common_keys(module_parameters)
        if {'K', 'L', 'n'} <= params:
            return self.physical_aoi_loss
        elif {'B5', 'B4', 'B3', 'B2', 'B1', 'B0'} <= params:
            return self.sapm_aoi_loss
        elif {'b'} <= params:
            return self.ashrae_aoi_loss
        elif {'a_r'} <= params:
            return self.martin_ruiz_aoi_loss
        else:
            raise ValueError('could not infer AOI model from '
                             'system.arrays[i].module_parameters. Check that '
                             'the module_parameters for all Arrays in '
                             'system.arrays contain parameters for '
                             'the physical, aoi, ashrae or martin_ruiz model; '
                             'explicitly set the model with the aoi_model '
                             'kwarg; or set aoi_model="no_loss".')

    def ashrae_aoi_loss(self):
        self.results.aoi_modifier = self.system.get_iam(
            self.results.aoi,
            iam_model='ashrae'
        )
        return self

    def physical_aoi_loss(self):
        self.results.aoi_modifier = self.system.get_iam(
            self.results.aoi,
            iam_model='physical'
        )
        return self

    def sapm_aoi_loss(self):
        self.results.aoi_modifier = self.system.get_iam(
            self.results.aoi,
            iam_model='sapm'
        )
        return self

    def martin_ruiz_aoi_loss(self):
        self.results.aoi_modifier = self.system.get_iam(
            self.results.aoi, iam_model='martin_ruiz'
        )
        return self

    def no_aoi_loss(self):
        if self.system.num_arrays == 1:
            self.results.aoi_modifier = 1.0
        else:
            self.results.aoi_modifier = (1.0,) * self.system.num_arrays
        return self

    @property
    def spectral_model(self):
        return self._spectral_model

    @spectral_model.setter
    def spectral_model(self, model):
        if model is None:
            self._spectral_model = self.infer_spectral_model()
        elif isinstance(model, str):
            model = model.lower()
            if model == 'first_solar':
                self._spectral_model = self.first_solar_spectral_loss
            elif model == 'sapm':
                self._spectral_model = self.sapm_spectral_loss
            elif model == 'no_loss':
                self._spectral_model = self.no_spectral_loss
            else:
                raise ValueError(model + ' is not a valid spectral loss model')
        else:
            self._spectral_model = partial(model, self)

    def infer_spectral_model(self):
        """Infer spectral model from system attributes."""
        module_parameters = tuple(
            array.module_parameters for array in self.system.arrays)
        params = _common_keys(module_parameters)
        if {'A4', 'A3', 'A2', 'A1', 'A0'} <= params:
            return self.sapm_spectral_loss
        elif ((('Technology' in params or
                'Material' in params) and
               (self.system._infer_cell_type() is not None)) or
              'first_solar_spectral_coefficients' in params):
            return self.first_solar_spectral_loss
        else:
            raise ValueError('could not infer spectral model from '
                             'system.arrays[i].module_parameters. Check that '
                             'the module_parameters for all Arrays in '
                             'system.arrays contain valid '
                             'first_solar_spectral_coefficients, a valid '
                             'Material or Technology value, or set '
                             'spectral_model="no_loss".')

    def first_solar_spectral_loss(self):
        self.results.spectral_modifier = self.system.first_solar_spectral_loss(
            _tuple_from_dfs(self.results.weather, 'precipitable_water'),
            self.results.airmass['airmass_absolute']
        )
        return self

    def sapm_spectral_loss(self):
        self.results.spectral_modifier = self.system.sapm_spectral_loss(
            self.results.airmass['airmass_absolute']
        )
        return self

    def no_spectral_loss(self):
        if self.system.num_arrays == 1:
            self.results.spectral_modifier = 1
        else:
            self.results.spectral_modifier = (1,) * self.system.num_arrays
        return self

    @property
    def temperature_model(self):
        return self._temperature_model

    @temperature_model.setter
    def temperature_model(self, model):
        if model is None:
            self._temperature_model = self.infer_temperature_model()
        elif isinstance(model, str):
            model = model.lower()
            if model == 'sapm':
                self._temperature_model = self.sapm_temp
            elif model == 'pvsyst':
                self._temperature_model = self.pvsyst_temp
            elif model == 'faiman':
                self._temperature_model = self.faiman_temp
            elif model == 'fuentes':
                self._temperature_model = self.fuentes_temp
            elif model == 'noct_sam':
                self._temperature_model = self.noct_sam_temp
            else:
                raise ValueError(model + ' is not a valid temperature model')
            # check system.temperature_model_parameters for consistency
            name_from_params = self.infer_temperature_model().__name__
            if self._temperature_model.__name__ != name_from_params:
                common_params = _common_keys(tuple(
                    array.temperature_model_parameters
                    for array in self.system.arrays))
                raise ValueError(
                    f'Temperature model {self._temperature_model.__name__} is '
                    f'inconsistent with PVSystem temperature model '
                    f'parameters. All Arrays in system.arrays must have '
                    f'consistent parameters. Common temperature model '
                    f'parameters: {common_params}'
                )
        else:
            self._temperature_model = partial(model, self)

    def infer_temperature_model(self):
        """Infer temperature model from system attributes."""
        temperature_model_parameters = tuple(
            array.temperature_model_parameters for array in self.system.arrays)
        params = _common_keys(temperature_model_parameters)
        # remove or statement in v0.9
        if {'a', 'b', 'deltaT'} <= params or (
                not params and self.system.racking_model is None
                and self.system.module_type is None):
            return self.sapm_temp
        elif {'u_c', 'u_v'} <= params:
            return self.pvsyst_temp
        elif {'u0', 'u1'} <= params:
            return self.faiman_temp
        elif {'noct_installed'} <= params:
            return self.fuentes_temp
        elif {'noct', 'module_efficiency'} <= params:
            return self.noct_sam_temp
        else:
            raise ValueError(f'could not infer temperature model from '
                             f'system.temperature_model_parameters. Check '
                             f'that all Arrays in system.arrays have '
                             f'parameters for the same temperature model. '
                             f'Common temperature model parameters: {params}.')

    def _set_celltemp(self, model):
        """Set self.results.cell_temperature using the given cell
        temperature model.

        Parameters
        ----------
        model : str
            A cell temperature model name to pass to
            :py:meth:`pvlib.pvsystem.PVSystem.get_cell_temperature`.
            Valid names are 'sapm', 'pvsyst', 'faiman', 'fuentes', 'noct_sam'

        Returns
        -------
        self
        """

        poa = _irrad_for_celltemp(self.results.total_irrad,
                                  self.results.effective_irradiance)
        temp_air = _tuple_from_dfs(self.results.weather, 'temp_air')
        wind_speed = _tuple_from_dfs(self.results.weather, 'wind_speed')
        kwargs = {}
        if model == 'noct_sam':
            kwargs['effective_irradiance'] = self.results.effective_irradiance
        self.results.cell_temperature = self.system.get_cell_temperature(
            poa, temp_air, wind_speed, model=model, **kwargs)
        return self

    def sapm_temp(self):
        return self._set_celltemp('sapm')

    def pvsyst_temp(self):
        return self._set_celltemp('pvsyst')

    def faiman_temp(self):
        return self._set_celltemp('faiman')

    def fuentes_temp(self):
        return self._set_celltemp('fuentes')

    def noct_sam_temp(self):
        return self._set_celltemp('noct_sam')

    @property
    def dc_ohmic_model(self):
        return self._dc_ohmic_model

    @dc_ohmic_model.setter
    def dc_ohmic_model(self, model):
        if isinstance(model, str):
            model = model.lower()
            if model == 'dc_ohms_from_percent':
                self._dc_ohmic_model = self.dc_ohms_from_percent
            elif model == 'no_loss':
                self._dc_ohmic_model = self.no_dc_ohmic_loss
            else:
                raise ValueError(model + ' is not a valid losses model')
        else:
            self._dc_ohmic_model = partial(model, self)

    def dc_ohms_from_percent(self):
        """
        Calculate time series of ohmic losses and apply those to the mpp power
        output of the `dc_model` based on the pvsyst equivalent resistance
        method. Uses a `dc_ohmic_percent` parameter in the `losses_parameters`
        of the PVsystem.
        """
        Rw = self.system.dc_ohms_from_percent()
        if isinstance(self.results.dc, tuple):
            self.results.dc_ohmic_losses = tuple(
                pvsystem.dc_ohmic_losses(Rw, df['i_mp'])
                for Rw, df in zip(Rw, self.results.dc)
            )
            for df, loss in zip(self.results.dc, self.results.dc_ohmic_losses):
                df['p_mp'] = df['p_mp'] - loss
        else:
            self.results.dc_ohmic_losses = pvsystem.dc_ohmic_losses(
                Rw, self.results.dc['i_mp']
            )
            self.results.dc['p_mp'] = (self.results.dc['p_mp']
                                       - self.results.dc_ohmic_losses)
        return self

    def no_dc_ohmic_loss(self):
        return self

    @property
    def losses_model(self):
        return self._losses_model

    @losses_model.setter
    def losses_model(self, model):
        if model is None:
            self._losses_model = self.infer_losses_model()
        elif isinstance(model, str):
            model = model.lower()
            if model == 'pvwatts':
                self._losses_model = self.pvwatts_losses
            elif model == 'no_loss':
                self._losses_model = self.no_extra_losses
            else:
                raise ValueError(model + ' is not a valid losses model')
        else:
            self._losses_model = partial(model, self)

    def infer_losses_model(self):
        raise NotImplementedError

    def pvwatts_losses(self):
        self.results.losses = (100 - self.system.pvwatts_losses()) / 100.
        if isinstance(self.results.dc, tuple):
            for dc in self.results.dc:
                dc *= self.results.losses
        else:
            self.results.dc *= self.results.losses
        return self

    def no_extra_losses(self):
        self.results.losses = 1
        return self

    def effective_irradiance_model(self):
        def _eff_irrad(module_parameters, total_irrad, spect_mod, aoi_mod):
            fd = module_parameters.get('FD', 1.)
            return spect_mod * (total_irrad['poa_direct'] * aoi_mod +
                                fd * total_irrad['poa_diffuse'])
        if isinstance(self.results.total_irrad, tuple):
            self.results.effective_irradiance = tuple(
                _eff_irrad(array.module_parameters, ti, sm, am) for
                array, ti, sm, am in zip(
                    self.system.arrays, self.results.total_irrad,
                    self.results.spectral_modifier, self.results.aoi_modifier))
        else:
            self.results.effective_irradiance = _eff_irrad(
                self.system.arrays[0].module_parameters,
                self.results.total_irrad,
                self.results.spectral_modifier,
                self.results.aoi_modifier
            )
        return self

    def complete_irradiance(self, weather):
        """
        Determine the missing irradiation columns. Only two of the
        following data columns (dni, ghi, dhi) are needed to calculate
        the missing data.

        This function is not safe at the moment. Results can be too high
        or negative. Please contribute and help to improve this function
        on https://github.com/pvlib/pvlib-python

        Parameters
        ----------
        weather : DataFrame, or tuple or list of DataFrame
            Column names must be ``'dni'``, ``'ghi'``, ``'dhi'``,
            ``'wind_speed'``, ``'temp_air'``. All irradiance components
            are required. Air temperature of 20 C and wind speed
            of 0 m/s will be added to the DataFrame if not provided.
            If `weather` is a tuple it must be the same length as the number
            of Arrays in the system and the indices for each DataFrame must
            be the same.

        Returns
        -------
        self

        Raises
        ------
        ValueError
            if the number of dataframes in `weather` is not the same as the
            number of Arrays in the system or if the indices of all elements
            of `weather` are not the same.

        Notes
        -----
        Assigns attributes to ``results``: ``times``, ``weather``

        Examples
        --------
        This example does not work until the parameters `my_system`,
        `my_location`, and `my_weather` are defined but shows the basic idea
        how this method can be used.

        >>> from pvlib.modelchain import ModelChain

        >>> # my_weather containing 'dhi' and 'ghi'.
        >>> mc = ModelChain(my_system, my_location)  # doctest: +SKIP
        >>> mc.complete_irradiance(my_weather)  # doctest: +SKIP
        >>> mc.run_model(mc.results.weather)  # doctest: +SKIP

        >>> # my_weather containing 'dhi', 'ghi' and 'dni'.
        >>> mc = ModelChain(my_system, my_location)  # doctest: +SKIP
        >>> mc.run_model(my_weather)  # doctest: +SKIP
        """
        weather = _to_tuple(weather)
        self._check_multiple_input(weather)
        # Don't use ModelChain._assign_weather() here because it adds
        # temperature and wind-speed columns which we do not need here.
        self.results.weather = _copy(weather)
        self._assign_times()
        self.results.solar_position = self.location.get_solarposition(
            self.results.times, method=self.solar_position_method)
        # Calculate the irradiance using the component sum equations,
        # if needed
        if isinstance(weather, tuple):
            for w in self.results.weather:
                self._complete_irradiance(w)
        else:
            self._complete_irradiance(self.results.weather)
        return self

    def _complete_irradiance(self, weather):
        icolumns = set(weather.columns)
        wrn_txt = ("This function is not safe at the moment.\n" +
                   "Results can be too high or negative.\n" +
                   "Help to improve this function on github:\n" +
                   "https://github.com/pvlib/pvlib-python \n")
        if {'ghi', 'dhi'} <= icolumns and 'dni' not in icolumns:
            clearsky = self.location.get_clearsky(
                weather.index, solar_position=self.results.solar_position)
            complete_irrad_df = pvlib.irradiance.complete_irradiance(
                solar_zenith=self.results.solar_position.zenith,
                ghi=weather.ghi,
                dhi=weather.dhi,
                dni=None,
                dni_clear=clearsky.dni)
            weather.loc[:, 'dni'] = complete_irrad_df.dni
        elif {'dni', 'dhi'} <= icolumns and 'ghi' not in icolumns:
            warnings.warn(wrn_txt, UserWarning)
            complete_irrad_df = pvlib.irradiance.complete_irradiance(
                solar_zenith=self.results.solar_position.zenith,
                ghi=None,
                dhi=weather.dhi,
                dni=weather.dni)
            weather.loc[:, 'ghi'] = complete_irrad_df.ghi
        elif {'dni', 'ghi'} <= icolumns and 'dhi' not in icolumns:
            warnings.warn(wrn_txt, UserWarning)
            complete_irrad_df = pvlib.irradiance.complete_irradiance(
                solar_zenith=self.results.solar_position.zenith,
                ghi=weather.ghi,
                dhi=None,
                dni=weather.dni)
            weather.loc[:, 'dhi'] = complete_irrad_df.dhi

    def _prep_inputs_solar_pos(self, weather):
        """
        Assign solar position
        """
        # build weather kwargs for solar position calculation
        kwargs = _build_kwargs(['pressure', 'temp_air'],
                               weather[0] if isinstance(weather, tuple)
                               else weather)
        try:
            kwargs['temperature'] = kwargs.pop('temp_air')
        except KeyError:
            pass

        self.results.solar_position = self.location.get_solarposition(
            self.results.times, method=self.solar_position_method,
            **kwargs)
        return self

    def _prep_inputs_albedo(self, weather):
        """
        Get albedo from weather
        """
        try:
            self.results.albedo = _tuple_from_dfs(weather, 'albedo')
        except KeyError:
            self.results.albedo = tuple([
                a.albedo for a in self.system.arrays])
        return self

    def _prep_inputs_airmass(self):
        """
        Assign airmass
        """
        self.results.airmass = self.location.get_airmass(
            solar_position=self.results.solar_position,
            model=self.airmass_model)
        return self

    def _prep_inputs_tracking(self):
        """
        Calculate tracker position and AOI
        """
        self.results.tracking = self.system.singleaxis(
            self.results.solar_position['apparent_zenith'],
            self.results.solar_position['azimuth'])
        self.results.tracking['surface_tilt'] = (
            self.results.tracking['surface_tilt']
                .fillna(self.system.axis_tilt))
        self.results.tracking['surface_azimuth'] = (
            self.results.tracking['surface_azimuth']
                .fillna(self.system.axis_azimuth))
        self.results.aoi = self.results.tracking['aoi']
        return self

    def _prep_inputs_fixed(self):
        """
        Calculate AOI for fixed tilt system
        """
        self.results.aoi = self.system.get_aoi(
            self.results.solar_position['apparent_zenith'],
            self.results.solar_position['azimuth'])
        return self

    def _verify_df(self, data, required):
        """ Checks data for column names in required

        Parameters
        ----------
        data : Dataframe
        required : List of str

        Raises
        ------
        ValueError if any of required are not in data.columns.
        """
        def _verify(data, index=None):
            if not set(required) <= set(data.columns):
                tuple_txt = "" if index is None else f"in element {index} "
                raise ValueError(
                    "Incomplete input data. Data needs to contain "
                    f"{required}. Detected data {tuple_txt}contains: "
                    f"{list(data.columns)}")
        if not isinstance(data, tuple):
            _verify(data)
        else:
            for (i, array_data) in enumerate(data):
                _verify(array_data, i)

    def _configure_results(self, per_array_data):
        """Configure the type used for per-array fields in
        ModelChainResult.

        If ``per_array_data`` is True and the number of arrays in the
        system is 1, then per-array results are stored as length-1
        tuples. This overrides the PVSystem defaults of unpacking a 1
        length tuple into a singleton.

        Parameters
        ----------
        per_array_data : bool
            If input data is provided for each array, pass True. If a
            single input data is provided for all arrays, pass False.
        """
        self.results._singleton_tuples = (
            self.system.num_arrays == 1 and per_array_data
        )

    def _assign_weather(self, data):
        def _build_weather(data):
            key_list = [k for k in WEATHER_KEYS if k in data]
            weather = data[key_list].copy()
            if weather.get('wind_speed') is None:
                weather['wind_speed'] = 0
            if weather.get('temp_air') is None:
                weather['temp_air'] = 20
            return weather
        if isinstance(data, tuple):
            weather = tuple(_build_weather(wx) for wx in data)
            self._configure_results(per_array_data=True)
        else:
            weather = _build_weather(data)
            self._configure_results(per_array_data=False)
        self.results.weather = weather
        self._assign_times()
        return self

    def _assign_total_irrad(self, data):
        def _build_irrad(data):
            key_list = [k for k in POA_KEYS if k in data]
            return data[key_list].copy()
        if isinstance(data, tuple):
            self.results.total_irrad = tuple(
                _build_irrad(irrad_data) for irrad_data in data
            )
            return self
        self.results.total_irrad = _build_irrad(data)
        return self

    def _assign_times(self):
        """Assign self.results.times according the the index of
        self.results.weather.

        If there are multiple DataFrames in self.results.weather then
        the index of the first one is assigned. It is assumed that the
        indices of each DataFrame in self.results.weather are the same.
        This can be verified by calling :py:func:`_all_same_index` or
        :py:meth:`self._check_multiple_weather` before calling this
        method.
        """
        if isinstance(self.results.weather, tuple):
            self.results.times = self.results.weather[0].index
        else:
            self.results.times = self.results.weather.index

    def prepare_inputs(self, weather):
        """
        Prepare the solar position, irradiance, and weather inputs to
        the model, starting with GHI, DNI and DHI.

        Parameters
        ----------
        weather : DataFrame, or tuple or list of DataFrames
            Required column names include ``'dni'``, ``'ghi'``, ``'dhi'``.
            Optional column names are ``'wind_speed'``, ``'temp_air'``,
            ``'albedo'``.

            If optional columns ``'wind_speed'``, ``'temp_air'`` are not
            provided, air temperature of 20 C and wind speed
            of 0 m/s will be added to the ``weather`` DataFrame.

            If optional column ``'albedo'`` is provided, albedo values in the
            ModelChain's PVSystem.arrays are ignored.

            If `weather` is a tuple or list, it must be of the same length and
            order as the Arrays of the ModelChain's PVSystem.

        Raises
        ------
        ValueError
            If any `weather` DataFrame(s) is missing an irradiance component.
        ValueError
            If `weather` is a tuple or list and the DataFrames it contains have
            different indices.
        ValueError
            If `weather` is a tuple or list with a different length than the
            number of Arrays in the system.

        Notes
        -----
        Assigns attributes to ``results``: ``times``, ``weather``,
        ``solar_position``, ``airmass``, ``total_irrad``, ``aoi``, ``albedo``.

        See also
        --------
        ModelChain.complete_irradiance
        """
        weather = _to_tuple(weather)
        self._check_multiple_input(weather, strict=False)
        self._verify_df(weather, required=['ghi', 'dni', 'dhi'])
        self._assign_weather(weather)

        self._prep_inputs_solar_pos(weather)
        self._prep_inputs_airmass()
        self._prep_inputs_albedo(weather)

        # PVSystem.get_irradiance and SingleAxisTracker.get_irradiance
        # and PVSystem.get_aoi and SingleAxisTracker.get_aoi
        # have different method signatures. Use partial to handle
        # the differences.
        if isinstance(self.system, SingleAxisTracker):
            self._prep_inputs_tracking()
            get_irradiance = partial(
                self.system.get_irradiance,
                self.results.tracking['surface_tilt'],
                self.results.tracking['surface_azimuth'],
                self.results.solar_position['apparent_zenith'],
                self.results.solar_position['azimuth'])
        else:
            self._prep_inputs_fixed()
            get_irradiance = partial(
                self.system.get_irradiance,
                self.results.solar_position['apparent_zenith'],
                self.results.solar_position['azimuth'])

        self.results.total_irrad = get_irradiance(
            _tuple_from_dfs(self.results.weather, 'dni'),
            _tuple_from_dfs(self.results.weather, 'ghi'),
            _tuple_from_dfs(self.results.weather, 'dhi'),
            albedo=self.results.albedo,
            airmass=self.results.airmass['airmass_relative'],
            model=self.transposition_model
        )

        return self

    def _check_multiple_input(self, data, strict=True):
        """Check that the number of elements in `data` is the same as
        the number of Arrays in `self.system`.

        In most cases if ``self.system.num_arrays`` is greater than 1 we
        want to raise an error when `data` is not a tuple; however, that
        behavior can be suppressed by setting ``strict=False``. This is
        useful for validating inputs such as GHI, DHI, DNI, wind speed, or
        air temperature that can be applied a ``PVSystem`` as a system-wide
        input. In this case we want to ensure that when a tuple is provided
        it has the same length as the number of Arrays, but we do not want
        to fail if the input is not a tuple.
        """
        if (not strict or self.system.num_arrays == 1) \
                and not isinstance(data, tuple):
            return
        if strict and not isinstance(data, tuple):
            raise TypeError("Input must be a tuple of length "
                            f"{self.system.num_arrays}, "
                            f"got {type(data).__name__}.")
        if len(data) != self.system.num_arrays:
            raise ValueError("Input must be same length as number of Arrays "
                             f"in system. Expected {self.system.num_arrays}, "
                             f"got {len(data)}.")
        _all_same_index(data)

    def prepare_inputs_from_poa(self, data):
        """
        Prepare the solar position, irradiance and weather inputs to
        the model, starting with plane-of-array irradiance.

        Parameters
        ----------
        data : DataFrame, or tuple or list of DataFrame
            Contains plane-of-array irradiance data. Required column names
            include ``'poa_global'``, ``'poa_direct'`` and ``'poa_diffuse'``.
            Columns with weather-related data are ssigned to the
            ``weather`` attribute.  If columns for ``'temp_air'`` and
            ``'wind_speed'`` are not provided, air temperature of 20 C and wind
            speed of 0 m/s are assumed.

            If list or tuple, must be of the same length and order as the
            Arrays of the ModelChain's PVSystem.

        Raises
        ------
        ValueError
             If the number of DataFrames passed in `data` is not the same
             as the number of Arrays in the system.

        Notes
        -----
        Assigns attributes to ``results``: ``times``, ``weather``,
        ``total_irrad``, ``solar_position``, ``airmass``, ``aoi``.

        See also
        --------
        pvlib.modelchain.ModelChain.prepare_inputs
        """
        data = _to_tuple(data)
        self._check_multiple_input(data)
        self._assign_weather(data)

        self._verify_df(data, required=['poa_global', 'poa_direct',
                                        'poa_diffuse'])
        self._assign_total_irrad(data)

        self._prep_inputs_solar_pos(data)
        self._prep_inputs_airmass()

        if isinstance(self.system, SingleAxisTracker):
            self._prep_inputs_tracking()
        else:
            self._prep_inputs_fixed()

        return self

    def _get_cell_temperature(self, data,
                              poa, temperature_model_parameters):
        """Extract the cell temperature data from a DataFrame.

        If 'cell_temperature' column exists in data then it is returned. If
        'module_temperature' column exists in data, then it is used with poa to
        calculate the cell temperature. If neither column exists then None is
        returned.

        Parameters
        ----------
        data : DataFrame (not a tuple of DataFrame)
        poa : Series (not a tuple of Series)

        Returns
        -------
        Series
        """
        if 'cell_temperature' in data:
            return data['cell_temperature']
        # cell_temperature is not in input. Calculate cell_temperature using
        # a temperature_model.
        # If module_temperature is in input data we can use the SAPM cell
        # temperature model.
        if (('module_temperature' in data) and
                (self.temperature_model == self.sapm_temp)):
            # use SAPM cell temperature model only
            return pvlib.temperature.sapm_cell_from_module(
                module_temperature=data['module_temperature'],
                poa_global=poa,
                deltaT=temperature_model_parameters['deltaT'])

    def _prepare_temperature_single_array(self, data, poa):
        """Set cell_temperature using a single data frame."""
        self.results.cell_temperature = self._get_cell_temperature(
            data,
            poa,
            self.system.arrays[0].temperature_model_parameters
        )
        if self.results.cell_temperature is None:
            self.temperature_model()
        return self

    def _prepare_temperature(self, data=None):
        """
        Sets cell_temperature using inputs in data and the specified
        temperature model.

        If 'data' contains 'cell_temperature', these values are assigned to
        attribute ``cell_temperature``. If 'data' contains 'module_temperature`
        and `temperature_model' is 'sapm', cell temperature is calculated using
        :py:func:`pvlib.temperature.sapm_cell_from_module`. Otherwise, cell
        temperature is calculated by 'temperature_model'.

        Parameters
        ----------
        data : DataFrame, default None
            May contain columns ``'cell_temperature'`` or
            ``'module_temperaure'``.

        Returns
        -------
        self

        Assigns attribute ``results.cell_temperature``.

        """
        poa = _irrad_for_celltemp(self.results.total_irrad,
                                  self.results.effective_irradiance)
        # handle simple case first, single array, data not iterable
        if not isinstance(data, tuple) and self.system.num_arrays == 1:
            return self._prepare_temperature_single_array(data, poa)
        if not isinstance(data, tuple):
            # broadcast data to all arrays
            data = (data,) * self.system.num_arrays
        # data is tuple, so temperature_model_parameters must also be
        # tuple. system.temperature_model_parameters is reduced to a dict
        # if system.num_arrays == 1, so manually access parameters. GH 1192
        t_mod_params = tuple(array.temperature_model_parameters
                             for array in self.system.arrays)
        # find where cell or module temperature is specified in input data
        given_cell_temperature = tuple(itertools.starmap(
            self._get_cell_temperature, zip(data, poa, t_mod_params)
        ))
        # If cell temperature has been specified for all arrays return
        # immediately and do not try to compute it.
        if all(cell_temp is not None for cell_temp in given_cell_temperature):
            self.results.cell_temperature = given_cell_temperature
            return self
        # Calculate cell temperature from weather data. If cell_temperature
        # has not been provided for some arrays then it is computed.
        self.temperature_model()
        # replace calculated cell temperature with temperature given in `data`
        # where available.
        self.results.cell_temperature = tuple(
            itertools.starmap(
                lambda given, modeled: modeled if given is None else given,
                zip(given_cell_temperature, self.results.cell_temperature)
            )
        )
        return self

    def run_model(self, weather):
        """
        Run the model chain starting with broadband global, diffuse and/or
        direct irradiance.

        Parameters
        ----------
        weather : DataFrame, or tuple or list of DataFrame
            Column names must include:

            - ``'dni'``
            - ``'ghi'``
            - ``'dhi'``

            Optional columns are:

            - ``'temp_air'``
            - ``'cell_temperature'``
            - ``'module_temperature'``
            - ``'wind_speed'``
            - ``'albedo'``

            If optional columns ``'temp_air'`` and ``'wind_speed'``
            are not provided, air temperature of 20 C and wind speed of 0 m/s
            are added to the DataFrame. If optional column
            ``'cell_temperature'`` is provided, these values are used instead
            of `temperature_model`. If optional column ``'module_temperature'``
            is provided, ``temperature_model`` must be ``'sapm'``.

            If optional column ``'albedo'`` is provided, ``'albedo'`` may not
            be present on the ModelChain's PVSystem.Arrays.

            If weather is a list or tuple, it must be of the same length and
            order as the Arrays of the ModelChain's PVSystem.

        Returns
        -------
        self

        Raises
        ------
        ValueError
            If the number of DataFrames in `data` is different than the number
            of Arrays in the PVSystem.
        ValueError
            If the DataFrames in `data` have different indexes.

        Notes
        -----
        Assigns attributes to ``results``: ``times``, ``weather``,
        ``solar_position``, ``airmass``, ``total_irrad``, ``aoi``,
        ``aoi_modifier``, ``spectral_modifier``, and
        ``effective_irradiance``, ``cell_temperature``, ``dc``, ``ac``,
        ``losses``, ``diode_params`` (if dc_model is a single diode
        model).

        See also
        --------
        pvlib.modelchain.ModelChain.run_model_from_poa
        pvlib.modelchain.ModelChain.run_model_from_effective_irradiance
        """
        weather = _to_tuple(weather)
        self.prepare_inputs(weather)
        self.aoi_model()
        self.spectral_model()
        self.effective_irradiance_model()

        self._run_from_effective_irrad(weather)

        return self

    def run_model_from_poa(self, data):
        """
        Run the model starting with broadband irradiance in the plane of array.

        Data must include direct, diffuse and total irradiance (W/m2) in the
        plane of array. Reflections and spectral adjustments are made to
        calculate effective irradiance (W/m2).

        Parameters
        ----------
        data : DataFrame, or tuple or list of DataFrame
            Required column names include ``'poa_global'``,
            ``'poa_direct'`` and ``'poa_diffuse'``. If optional columns
            ``'temp_air'`` and ``'wind_speed'`` are not provided, air
            temperature of 20 C and wind speed of 0 m/s are assumed.
            If optional column ``'cell_temperature'`` is provided, these values
            are used instead of `temperature_model`. If optional column
            ``'module_temperature'`` is provided, `temperature_model` must be
            ``'sapm'``.

            If the ModelChain's PVSystem has multiple arrays, `data` must be a
            list or tuple with the same length and order as the PVsystem's
            Arrays. Each element of `data` provides the irradiance and weather
            for the corresponding array.

        Returns
        -------
        self

        Raises
        ------
        ValueError
            If the number of DataFrames in `data` is different than the number
            of Arrays in the PVSystem.
        ValueError
            If the DataFrames in `data` have different indexes.

        Notes
        -----
        Assigns attributes to results: ``times``, ``weather``,
        ``solar_position``, ``airmass``, ``total_irrad``, ``aoi``,
        ``aoi_modifier``, ``spectral_modifier``, and
        ``effective_irradiance``, ``cell_temperature``, ``dc``, ``ac``,
        ``losses``, ``diode_params`` (if dc_model is a single diode
        model).

        See also
        --------
        pvlib.modelchain.ModelChain.run_model
        pvlib.modelchain.ModelChain.run_model_from_effective_irradiance
        """
        data = _to_tuple(data)
        self.prepare_inputs_from_poa(data)

        self.aoi_model()
        self.spectral_model()
        self.effective_irradiance_model()

        self._run_from_effective_irrad(data)

        return self

    def _run_from_effective_irrad(self, data=None):
        """
        Executes the temperature, DC, losses and AC models.

        Parameters
        ----------
        data : DataFrame, or tuple of DataFrame, default None
            If optional column ``'cell_temperature'`` is provided, these values
            are used instead of `temperature_model`. If optional column
            `module_temperature` is provided, `temperature_model` must be
            ``'sapm'``.

        Returns
        -------
        self

        Notes
        -----
        Assigns attributes:``cell_temperature``, ``dc``, ``ac``, ``losses``,
        ``diode_params`` (if dc_model is a single diode model).
        """
        self._prepare_temperature(data)
        self.dc_model()
        self.dc_ohmic_model()
        self.losses_model()
        self.ac_model()

        return self

    def run_model_from_effective_irradiance(self, data=None):
        """
        Run the model starting with effective irradiance in the plane of array.

        Effective irradiance is irradiance in the plane-of-array after any
        adjustments for soiling, reflections and spectrum.

        Parameters
        ----------
        data : DataFrame, or list or tuple of DataFrame
            Required column is ``'effective_irradiance'``.
            Optional columns include ``'cell_temperature'``,
            ``'module_temperature'`` and ``'poa_global'``.

            If the ModelChain's PVSystem has multiple arrays, `data` must be a
            list or tuple with the same length and order as the PVsystem's
            Arrays. Each element of `data` provides the irradiance and weather
            for the corresponding array.

        Returns
        -------
        self

        Raises
        ------
        ValueError
            If the number of DataFrames in `data` is different than the number
            of Arrays in the PVSystem.
        ValueError
            If the DataFrames in `data` have different indexes.

        Notes
        -----
        Optional ``data`` columns ``'cell_temperature'``,
        ``'module_temperature'`` and ``'poa_global'`` are used for determining
        cell temperature.

        * If optional column ``'cell_temperature'`` is present, these values
          are used and `temperature_model` is ignored.
        * If optional column ``'module_temperature'`` is preset,
          `temperature_model` must be ``'sapm'``.
        * Otherwise, cell temperature is calculated using `temperature_model`.

        The cell temperature models require plane-of-array irradiance as input.
        If optional column ``'poa_global'`` is present, these data are used.
        If ``'poa_global'`` is not present, ``'effective_irradiance'`` is used.

        Assigns attributes to results: ``times``, ``weather``, ``total_irrad``,
        ``effective_irradiance``, ``cell_temperature``, ``dc``, ``ac``,
        ``losses``, ``diode_params`` (if dc_model is a single diode model).

        See also
        --------
        pvlib.modelchain.ModelChain.run_model
        pvlib.modelchain.ModelChain.run_model_from_poa
        """
        data = _to_tuple(data)
        self._check_multiple_input(data)
        self._verify_df(data, required=['effective_irradiance'])
        self._assign_weather(data)
        self._assign_total_irrad(data)
        self.results.effective_irradiance = _tuple_from_dfs(
            data, 'effective_irradiance')
        self._run_from_effective_irrad(data)

        return self


def _irrad_for_celltemp(total_irrad, effective_irradiance):
    """
    Determine irradiance to use for cell temperature models, in order
    of preference 'poa_global' then 'effective_irradiance'

    Returns
    -------
    Series or tuple of Series
        tuple if total_irrad is a tuple of DataFrame

    """
    if isinstance(total_irrad, tuple):
        if all(['poa_global' in df for df in total_irrad]):
            return _tuple_from_dfs(total_irrad, 'poa_global')
        else:
            return effective_irradiance
    else:
        if 'poa_global' in total_irrad:
            return total_irrad['poa_global']
        else:
            return effective_irradiance


def _snl_params(inverter_params):
    """Return True if `inverter_params` includes parameters for the
    Sandia inverter model."""
    return {'C0', 'C1', 'C2'} <= inverter_params


def _adr_params(inverter_params):
    """Return True if `inverter_params` includes parameters for the ADR
    inverter model."""
    return {'ADRCoefficients'} <= inverter_params


def _pvwatts_params(inverter_params):
    """Return True if `inverter_params` includes parameters for the
    PVWatts inverter model."""
    return {'pdc0'} <= inverter_params


def _copy(data):
    """Return a copy of each DataFrame in `data` if it is a tuple,
    otherwise return a copy of `data`."""
    if not isinstance(data, tuple):
        return data.copy()
    return tuple(df.copy() for df in data)


def _all_same_index(data):
    """Raise a ValueError if all DataFrames in `data` do not have the
    same index."""
    indexes = map(lambda df: df.index, data)
    next(indexes, None)
    for index in indexes:
        if not index.equals(data[0].index):
            raise ValueError("Input DataFrames must have same index.")


def _common_keys(dicts):
    """Return the intersection of the set of keys for each dictionary
    in `dicts`"""
    def _keys(x):
        return set(x.keys())
    if isinstance(dicts, tuple):
        return set.intersection(*map(_keys, dicts))
    return _keys(dicts)


def _tuple_from_dfs(dfs, name):
    """Extract a column from each DataFrame in `dfs` if `dfs` is a tuple.

    Returns a tuple of Series if `dfs` is a tuple or a Series if `dfs` is
    a DataFrame.
    """
    if isinstance(dfs, tuple):
        return tuple(df[name] for df in dfs)
    else:
        return dfs[name]


def _to_tuple(x):
    if not isinstance(x, (tuple, list)):
        return x
    return tuple(x)



"""
Collection of functions used in pvlib_python
"""

import datetime as dt
import numpy as np
import pandas as pd
import pytz


def cosd(angle):
    """
    Cosine with angle input in degrees

    Parameters
    ----------
    angle : float or array-like
        Angle in degrees

    Returns
    -------
    result : float or array-like
        Cosine of the angle
    """

    res = np.cos(np.radians(angle))
    return res


def sind(angle):
    """
    Sine with angle input in degrees

    Parameters
    ----------
    angle : float
        Angle in degrees

    Returns
    -------
    result : float
        Sin of the angle
    """

    res = np.sin(np.radians(angle))
    return res


def tand(angle):
    """
    Tan with angle input in degrees

    Parameters
    ----------
    angle : float
        Angle in degrees

    Returns
    -------
    result : float
        Tan of the angle
    """

    res = np.tan(np.radians(angle))
    return res


def asind(number):
    """
    Inverse Sine returning an angle in degrees

    Parameters
    ----------
    number : float
        Input number

    Returns
    -------
    result : float
        arcsin result
    """

    res = np.degrees(np.arcsin(number))
    return res


def localize_to_utc(time, location):
    """
    Converts or localizes a time series to UTC.

    Parameters
    ----------
    time : datetime.datetime, pandas.DatetimeIndex,
           or pandas.Series/DataFrame with a DatetimeIndex.
    location : pvlib.Location object

    Returns
    -------
    pandas object localized to UTC.
    """
    if isinstance(time, dt.datetime):
        if time.tzinfo is None:
            time = pytz.timezone(location.tz).localize(time)
        time_utc = time.astimezone(pytz.utc)
    else:
        try:
            time_utc = time.tz_convert('UTC')
        except TypeError:
            time_utc = time.tz_localize(location.tz).tz_convert('UTC')

    return time_utc


def datetime_to_djd(time):
    """
    Converts a datetime to the Dublin Julian Day

    Parameters
    ----------
    time : datetime.datetime
        time to convert

    Returns
    -------
    float
        fractional days since 12/31/1899+0000
    """

    if time.tzinfo is None:
        time_utc = pytz.utc.localize(time)
    else:
        time_utc = time.astimezone(pytz.utc)

    djd_start = pytz.utc.localize(dt.datetime(1899, 12, 31, 12))
    djd = (time_utc - djd_start).total_seconds() * 1.0/(60 * 60 * 24)

    return djd


def djd_to_datetime(djd, tz='UTC'):
    """
    Converts a Dublin Julian Day float to a datetime.datetime object

    Parameters
    ----------
    djd : float
        fractional days since 12/31/1899+0000
    tz : str, default 'UTC'
        timezone to localize the result to

    Returns
    -------
    datetime.datetime
       The resultant datetime localized to tz
    """

    djd_start = pytz.utc.localize(dt.datetime(1899, 12, 31, 12))

    utc_time = djd_start + dt.timedelta(days=djd)
    return utc_time.astimezone(pytz.timezone(tz))


def _pandas_to_doy(pd_object):
    """
    Finds the day of year for a pandas datetime-like object.

    Useful for delayed evaluation of the dayofyear attribute.

    Parameters
    ----------
    pd_object : DatetimeIndex or Timestamp

    Returns
    -------
    dayofyear
    """
    return pd_object.dayofyear


def _doy_to_datetimeindex(doy, epoch_year=2014):
    """
    Convert a day of year scalar or array to a pd.DatetimeIndex.

    Parameters
    ----------
    doy : numeric
        Contains days of the year

    Returns
    -------
    pd.DatetimeIndex
    """
    doy = np.atleast_1d(doy).astype('float')
    epoch = pd.Timestamp('{}-12-31'.format(epoch_year - 1))
    timestamps = [epoch + dt.timedelta(days=adoy) for adoy in doy]
    return pd.DatetimeIndex(timestamps)


def _datetimelike_scalar_to_doy(time):
    return pd.DatetimeIndex([pd.Timestamp(time)]).dayofyear


def _datetimelike_scalar_to_datetimeindex(time):
    return pd.DatetimeIndex([pd.Timestamp(time)])


def _scalar_out(arg):
    if np.isscalar(arg):
        output = arg
    else:  #
        # works if it's a 1 length array and
        # will throw a ValueError otherwise
        output = np.asarray(arg).item()

    return output


def _array_out(arg):
    if isinstance(arg, pd.Series):
        output = arg.values
    else:
        output = arg

    return output


def _build_kwargs(keys, input_dict):
    """
    Parameters
    ----------
    keys : iterable
        Typically a list of strings.
    adict : dict-like
        A dictionary from which to attempt to pull each key.

    Returns
    -------
    kwargs : dict
        A dictionary with only the keys that were in input_dict
    """

    kwargs = {}
    for key in keys:
        try:
            kwargs[key] = input_dict[key]
        except KeyError:
            pass

    return kwargs


# Created April,2014
# Author: Rob Andrews, Calama Consulting

def _golden_sect_DataFrame(params, VL, VH, func):
    """
    Vectorized golden section search for finding MPP from a dataframe
    timeseries.

    Parameters
    ----------
    params : dict
        Dictionary containing scalars or arrays
        of inputs to the function to be optimized.
        Each row should represent an independent optimization.

    VL: float
        Lower bound of the optimization

    VH: float
        Upper bound of the optimization

    func: function
        Function to be optimized must be in the form f(array-like, x)

    Returns
    -------
    func(df,'V1') : DataFrame
        function evaluated at the optimal point

    df['V1']: Dataframe
        Dataframe of optimal points

    Notes
    -----
    This function will find the MAXIMUM of a function
    """

    df = params
    df['VH'] = VH
    df['VL'] = VL

    errflag = True
    iterations = 0

    while errflag:

        phi = (np.sqrt(5)-1)/2*(df['VH']-df['VL'])
        df['V1'] = df['VL'] + phi
        df['V2'] = df['VH'] - phi

        df['f1'] = func(df, 'V1')
        df['f2'] = func(df, 'V2')
        df['SW_Flag'] = df['f1'] > df['f2']

        df['VL'] = df['V2']*df['SW_Flag'] + df['VL']*(~df['SW_Flag'])
        df['VH'] = df['V1']*~df['SW_Flag'] + df['VH']*(df['SW_Flag'])

        err = df['V1'] - df['V2']
        try:
            errflag = (abs(err) > .01).any()
        except ValueError:
            errflag = (abs(err) > .01)

        iterations += 1

        if iterations > 50:
            raise Exception("EXCEPTION:iterations exceeded maximum (50)")

    return func(df, 'V1'), df['V1']



"""Implementation of Rule L011."""
from typing import List, Optional, Union

from sqlfluff.core.parser import (
    WhitespaceSegment,
    KeywordSegment,
)

from sqlfluff.core.rules.base import BaseRule, LintResult, LintFix, RuleContext
from sqlfluff.core.rules.doc_decorators import document_fix_compatible


@document_fix_compatible
class Rule_L011(BaseRule):
    """Implicit/explicit aliasing of table.

    Aliasing of table to follow preference
    (explicit using an `AS` clause is default).

    | **Anti-pattern**
    | In this example, the alias 'voo' is implicit.

    .. code-block:: sql

        SELECT
            voo.a
        FROM foo voo

    | **Best practice**
    | Add `AS` to make it explicit.

    .. code-block:: sql

        SELECT
            voo.a
        FROM foo AS voo

    """

    config_keywords = ["aliasing"]

    _target_elems = ("from_expression_element",)

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Implicit aliasing of table/column not allowed. Use explicit `AS` clause.

        We look for the alias segment, and then evaluate its parent and whether
        it contains an AS keyword. This is the _eval function for both L011 and L012.

        The use of `raw_stack` is just for working out how much whitespace to add.

        """
        fixes = []

        if context.segment.is_type("alias_expression"):
            if context.parent_stack[-1].is_type(*self._target_elems):
                if any(e.name.lower() == "as" for e in context.segment.segments):
                    if self.aliasing == "implicit":  # type: ignore
                        if context.segment.segments[0].name.lower() == "as":

                            # Remove the AS as we're using implict aliasing
                            fixes.append(LintFix.delete(context.segment.segments[0]))
                            anchor = context.raw_stack[-1]

                            # Remove whitespace before (if exists) or after (if not)
                            if (
                                len(context.raw_stack) > 0
                                and context.raw_stack[-1].type == "whitespace"
                            ):
                                fixes.append(LintFix.delete(context.raw_stack[-1]))
                            elif (
                                len(context.segment.segments) > 0
                                and context.segment.segments[1].type == "whitespace"
                            ):
                                fixes.append(
                                    LintFix.delete(context.segment.segments[1])
                                )

                            return LintResult(anchor=anchor, fixes=fixes)

                else:
                    insert_buff: List[Union[WhitespaceSegment, KeywordSegment]] = []

                    # Add initial whitespace if we need to...
                    if context.raw_stack[-1].name not in ["whitespace", "newline"]:
                        insert_buff.append(WhitespaceSegment())

                    # Add an AS (Uppercase for now, but could be corrected later)
                    insert_buff.append(KeywordSegment("AS"))

                    # Add a trailing whitespace if we need to
                    if context.segment.segments[0].name not in [
                        "whitespace",
                        "newline",
                    ]:
                        insert_buff.append(WhitespaceSegment())

                    return LintResult(
                        anchor=context.segment,
                        fixes=[
                            LintFix.create_before(
                                context.segment.segments[0],
                                insert_buff,
                            )
                        ],
                    )
        return None



"""Heuristic-based template slicing algorithm.

This is the original slicing algorithm. It has limitations that may surface
when templates include similar strings multiple places in the file.
"""

import logging
import regex
from typing import Iterator, List

from jinja2 import Environment

from sqlfluff.core.templaters.base import RawFileSlice


# Instantiate the templater logger
templater_logger = logging.getLogger("sqlfluff.templater")


re_open_tag = regex.compile(r"^\s*{%[\+\-]?\s*")
re_close_tag = regex.compile(r"\s*[\+\-]?%}\s*$")


def slice_template(in_str: str, env: Environment) -> List[RawFileSlice]:
    """Slice template in jinja."""
    return list(_slice_template(in_str, env))


def _slice_template(in_str: str, env: Environment) -> Iterator[RawFileSlice]:
    """Helper function for slice template in jinja.

    NB: Starts and ends of blocks are not distinguished.
    """
    str_buff = ""
    idx = 0
    # We decide the "kind" of element we're dealing with
    # using it's _closing_ tag rather than it's opening
    # tag. The types here map back to similar types of
    # sections in the python slicer.
    block_types = {
        "variable_end": "templated",
        "block_end": "block",
        "comment_end": "comment",
        # Raw tags should behave like blocks. Note that
        # raw_end and raw_begin are whole tags rather
        # than blocks and comments where we get partial
        # tags.
        "raw_end": "block",
        "raw_begin": "block",
    }

    # https://jinja.palletsprojects.com/en/2.11.x/api/#jinja2.Environment.lex
    for _, elem_type, raw in env.lex(in_str):
        if elem_type == "data":
            yield RawFileSlice(raw, "literal", idx)
            idx += len(raw)
        else:  # pragma: no cover
            str_buff += raw

            if elem_type.endswith("_begin"):
                # When a "begin" tag (whether block, comment, or data) uses
                # whitespace stripping
                # (https://jinja.palletsprojects.com/en/3.0.x/templates/#whitespace-control),
                # the Jinja lex() function handles this by discarding adjacent
                # whitespace from in_str. For more insight, see the tokeniter()
                # function in this file:
                # https://github.com/pallets/jinja/blob/main/src/jinja2/lexer.py
                # We want to detect and correct for this in order to:
                # - Correctly update "idx" (if this is wrong, that's a
                #   potential DISASTER because lint fixes use this info to
                #   update the source file, and incorrect values often result in
                #   CORRUPTING the user's file so it's no longer valid SQL. :-O
                # - Guarantee that the slices we return fully "cover" the
                #   contents of in_str.
                #
                # We detect skipped characters by looking ahead in in_str for
                # the token just returned from lex(). The token text will either
                # be at the current 'idx' position (if whitespace stripping did
                # not occur) OR it'll be farther along in in_str, but we're
                # GUARANTEED that lex() only skips over WHITESPACE; nothing else.

                # Find the token returned. Did lex() skip over any characters?
                num_chars_skipped = in_str.index(raw, idx) - idx
                if num_chars_skipped:
                    # Yes. It skipped over some characters. Compute a string
                    # containing the skipped characters.
                    skipped_str = in_str[idx : idx + num_chars_skipped]

                    # Sanity check: Verify that Jinja only skips over
                    # WHITESPACE, never anything else.
                    if not skipped_str.isspace():  # pragma: no cover
                        templater_logger.warning(
                            "Jinja lex() skipped non-whitespace: %s", skipped_str
                        )
                    # Treat the skipped whitespace as a literal.
                    yield RawFileSlice(skipped_str, "literal", idx)
                    idx += num_chars_skipped

            # raw_end and raw_begin behave a little differently in
            # that the whole tag shows up in one go rather than getting
            # parts of the tag at a time.
            if elem_type.endswith("_end") or elem_type == "raw_begin":
                block_type = block_types[elem_type]
                block_subtype = None
                # Handle starts and ends of blocks
                if block_type == "block":
                    # Trim off the brackets and then the whitespace
                    m_open = re_open_tag.search(str_buff)
                    m_close = re_close_tag.search(str_buff)
                    trimmed_content = ""
                    if m_open and m_close:
                        trimmed_content = str_buff[
                            len(m_open.group(0)) : -len(m_close.group(0))
                        ]
                    if trimmed_content.startswith("end"):
                        block_type = "block_end"
                    elif trimmed_content.startswith("el"):
                        # else, elif
                        block_type = "block_mid"  # pragma: no cover
                    else:
                        block_type = "block_start"
                        if trimmed_content.split()[0] == "for":
                            block_subtype = "loop"  # pragma: no cover
                m = regex.search(r"\s+$", raw, regex.MULTILINE | regex.DOTALL)
                if raw.startswith("-") and m:
                    # Right whitespace was stripped. Split off the trailing
                    # whitespace into a separate slice. The desired behavior is
                    # to behave similarly as the left stripping case above.
                    # Note that the stakes are a bit different, because lex()
                    # hasn't *omitted* any characters from the strings it
                    # returns, it has simply grouped them differently than we
                    # want.
                    trailing_chars = len(m.group(0))
                    yield RawFileSlice(
                        str_buff[:-trailing_chars], block_type, idx, block_subtype
                    )
                    idx += len(str_buff) - trailing_chars
                    yield RawFileSlice(str_buff[-trailing_chars:], "literal", idx)
                    idx += trailing_chars
                else:
                    yield RawFileSlice(str_buff, block_type, idx, block_subtype)
                    idx += len(str_buff)
                str_buff = ""



"""
Created on Wed Oct  2 10:14:16 2019

@author: cwhanse
"""

import numpy as np
import pandas as pd

import pytest
from .conftest import assert_series_equal
from numpy.testing import assert_allclose

from pvlib import iam as _iam


def test_ashrae():
    thetas = np.array([-90., -67.5, -45., -22.5, 0., 22.5, 45., 67.5, 89., 90.,
                       np.nan])
    expected = np.array([0, 0.9193437, 0.97928932, 0.99588039, 1., 0.99588039,
                         0.97928932, 0.9193437, 0, 0, np.nan])
    iam = _iam.ashrae(thetas, .05)
    assert_allclose(iam, expected, equal_nan=True)
    iam_series = _iam.ashrae(pd.Series(thetas))
    assert_series_equal(iam_series, pd.Series(expected))


def test_ashrae_scalar():
    thetas = -45.
    iam = _iam.ashrae(thetas, .05)
    expected = 0.97928932
    assert_allclose(iam, expected, equal_nan=True)
    thetas = np.nan
    iam = _iam.ashrae(thetas, .05)
    expected = np.nan
    assert_allclose(iam, expected, equal_nan=True)


def test_physical():
    aoi = np.array([-90., -67.5, -45., -22.5, 0., 22.5, 45., 67.5, 90.,
                    np.nan])
    expected = np.array([0, 0.8893998, 0.98797788, 0.99926198, 1, 0.99926198,
                         0.98797788, 0.8893998, 0, np.nan])
    iam = _iam.physical(aoi, 1.526, 0.002, 4)
    assert_allclose(iam, expected, atol=1e-7, equal_nan=True)

    # GitHub issue 397
    aoi = pd.Series(aoi)
    iam = _iam.physical(aoi, 1.526, 0.002, 4)
    expected = pd.Series(expected)
    assert_series_equal(iam, expected)


def test_physical_n1_L0():
    aoi = np.array([0, 22.5, 45, 67.5, 90, 100, np.nan])
    expected = np.array([1, 1, 1, 1, 0, 0, np.nan])
    iam = _iam.physical(aoi, n=1, L=0)
    assert_allclose(iam, expected, equal_nan=True)

    aoi = pd.Series(aoi)
    expected = pd.Series(expected)
    iam = _iam.physical(aoi, n=1, L=0)
    assert_series_equal(iam, expected)


def test_physical_ar():
    aoi = np.array([0, 22.5, 45, 67.5, 90, 100, np.nan])
    expected = np.array([1, 0.99944171, 0.9917463, 0.91506158, 0, 0, np.nan])
    iam = _iam.physical(aoi, n_ar=1.29)
    assert_allclose(iam, expected, atol=1e-7, equal_nan=True)


def test_physical_noar():
    aoi = np.array([0, 22.5, 45, 67.5, 90, 100, np.nan])
    expected = _iam.physical(aoi)
    iam0 = _iam.physical(aoi, n_ar=1)
    iam1 = _iam.physical(aoi, n_ar=1.526)
    assert_allclose(iam0, expected, equal_nan=True)
    assert_allclose(iam1, expected, equal_nan=True)


def test_physical_scalar():
    aoi = -45.
    iam = _iam.physical(aoi, 1.526, 0.002, 4)
    expected = 0.98797788
    assert_allclose(iam, expected, equal_nan=True)
    aoi = np.nan
    iam = _iam.physical(aoi, 1.526, 0.002, 4)
    expected = np.nan
    assert_allclose(iam, expected, equal_nan=True)


def test_martin_ruiz():

    aoi = 45.
    a_r = 0.16
    expected = 0.98986965

    # will fail if default values change
    iam = _iam.martin_ruiz(aoi)
    assert_allclose(iam, expected)

    # will fail if parameter names change
    iam = _iam.martin_ruiz(aoi=aoi, a_r=a_r)
    assert_allclose(iam, expected)

    a_r = 0.18
    aoi = [-100, -60, 0, 60, 100, np.nan, np.inf]
    expected = [0.0, 0.9414631, 1.0, 0.9414631, 0.0, np.nan, 0.0]

    # check out of range of inputs as list
    iam = _iam.martin_ruiz(aoi, a_r)
    assert_allclose(iam, expected, equal_nan=True)

    # check out of range of inputs as array
    iam = _iam.martin_ruiz(np.array(aoi), a_r)
    assert_allclose(iam, expected, equal_nan=True)

    # check out of range of inputs as Series
    aoi = pd.Series(aoi)
    expected = pd.Series(expected)
    iam = _iam.martin_ruiz(aoi, a_r)
    assert_series_equal(iam, expected)


def test_martin_ruiz_exception():

    with pytest.raises(ValueError):
        _iam.martin_ruiz(0.0, a_r=0.0)


def test_martin_ruiz_diffuse():

    surface_tilt = 30.
    a_r = 0.16
    expected = (0.9549735, 0.7944426)

    # will fail if default values change
    iam = _iam.martin_ruiz_diffuse(surface_tilt)
    assert_allclose(iam, expected)

    # will fail if parameter names change
    iam = _iam.martin_ruiz_diffuse(surface_tilt=surface_tilt, a_r=a_r)
    assert_allclose(iam, expected)

    a_r = 0.18
    surface_tilt = [0, 30, 90, 120, 180, np.nan, np.inf]
    expected_sky = [0.9407678, 0.9452250, 0.9407678, 0.9055541, 0.0000000,
                    np.nan, np.nan]
    expected_gnd = [0.0000000, 0.7610849, 0.9407678, 0.9483508, 0.9407678,
                    np.nan, np.nan]

    # check various inputs as list
    iam = _iam.martin_ruiz_diffuse(surface_tilt, a_r)
    assert_allclose(iam[0], expected_sky, atol=1e-7, equal_nan=True)
    assert_allclose(iam[1], expected_gnd, atol=1e-7, equal_nan=True)

    # check various inputs as array
    iam = _iam.martin_ruiz_diffuse(np.array(surface_tilt), a_r)
    assert_allclose(iam[0], expected_sky, atol=1e-7, equal_nan=True)
    assert_allclose(iam[1], expected_gnd, atol=1e-7, equal_nan=True)

    # check various inputs as Series
    surface_tilt = pd.Series(surface_tilt)
    expected_sky = pd.Series(expected_sky, name='iam_sky')
    expected_gnd = pd.Series(expected_gnd, name='iam_ground')
    iam = _iam.martin_ruiz_diffuse(surface_tilt, a_r)
    assert_series_equal(iam[0], expected_sky)
    assert_series_equal(iam[1], expected_gnd)


def test_iam_interp():

    aoi_meas = [0.0, 45.0, 65.0, 75.0]
    iam_meas = [1.0,  0.9,  0.8,  0.6]

    # simple default linear method
    aoi = 55.0
    expected = 0.85
    iam = _iam.interp(aoi, aoi_meas, iam_meas)
    assert_allclose(iam, expected)

    # simple non-default method
    aoi = 55.0
    expected = 0.8878062
    iam = _iam.interp(aoi, aoi_meas, iam_meas, method='cubic')
    assert_allclose(iam, expected)

    # check with all reference values
    aoi = aoi_meas
    expected = iam_meas
    iam = _iam.interp(aoi, aoi_meas, iam_meas)
    assert_allclose(iam, expected)

    # check normalization and Series
    aoi = pd.Series(aoi)
    expected = pd.Series(expected)
    iam_mult = np.multiply(0.9, iam_meas)
    iam = _iam.interp(aoi, aoi_meas, iam_mult, normalize=True)
    assert_series_equal(iam, expected)

    # check beyond reference values
    aoi = [-45, 0, 45, 85, 90, 95, 100, 105, 110]
    expected = [0.9, 1.0, 0.9, 0.4, 0.3, 0.2, 0.1, 0.0, 0.0]
    iam = _iam.interp(aoi, aoi_meas, iam_meas)
    assert_allclose(iam, expected)

    # check exception clause
    with pytest.raises(ValueError):
        _iam.interp(0.0, [0], [1])

    # check exception clause
    with pytest.raises(ValueError):
        _iam.interp(0.0, [0, 90], [1, -1])


@pytest.mark.parametrize('aoi,expected', [
    (45, 0.9975036250000002),
    (np.array([[-30, 30, 100, np.nan]]),
     np.array([[0, 1.007572, 0, np.nan]])),
    (pd.Series([80]), pd.Series([0.597472]))
])
def test_sapm(sapm_module_params, aoi, expected):

    out = _iam.sapm(aoi, sapm_module_params)

    if isinstance(aoi, pd.Series):
        assert_series_equal(out, expected, check_less_precise=4)
    else:
        assert_allclose(out, expected, atol=1e-4)


def test_sapm_limits():
    module_parameters = {'B0': 5, 'B1': 0, 'B2': 0, 'B3': 0, 'B4': 0, 'B5': 0}
    assert _iam.sapm(1, module_parameters) == 5

    module_parameters = {'B0': 5, 'B1': 0, 'B2': 0, 'B3': 0, 'B4': 0, 'B5': 0}
    assert _iam.sapm(1, module_parameters, upper=1) == 1

    module_parameters = {'B0': -5, 'B1': 0, 'B2': 0, 'B3': 0, 'B4': 0, 'B5': 0}
    assert _iam.sapm(1, module_parameters) == 0


def test_marion_diffuse_model(mocker):
    # 1: return values are correct
    # 2: the underlying models are called appropriately
    ashrae_expected = {
        'sky': 0.9596085829811408,
        'horizon': 0.8329070417832541,
        'ground': 0.719823559106309
    }
    physical_expected = {
        'sky': 0.9539178294437575,
        'horizon': 0.7652650139134007,
        'ground': 0.6387140117795903
    }
    ashrae_spy = mocker.spy(_iam, 'ashrae')
    physical_spy = mocker.spy(_iam, 'physical')

    ashrae_actual = _iam.marion_diffuse('ashrae', 20)
    assert ashrae_spy.call_count == 3  # one call for each of the 3 regions
    assert physical_spy.call_count == 0
    physical_actual = _iam.marion_diffuse('physical', 20)
    assert ashrae_spy.call_count == 3
    assert physical_spy.call_count == 3

    for k, v in ashrae_expected.items():
        np.testing.assert_allclose(ashrae_actual[k], v)

    for k, v in physical_expected.items():
        np.testing.assert_allclose(physical_actual[k], v)


def test_marion_diffuse_kwargs():
    # kwargs get passed to underlying model
    expected = {
        'sky': 0.967489994422575,
        'horizon': 0.8647842827418412,
        'ground': 0.7700443455928433
    }
    actual = _iam.marion_diffuse('ashrae', 20, b=0.04)

    for k, v in expected.items():
        np.testing.assert_allclose(actual[k], v)


def test_marion_diffuse_invalid():
    with pytest.raises(ValueError):
        _iam.marion_diffuse('not_a_model', 20)


@pytest.mark.parametrize('region,N,expected', [
    ('sky', 180, 0.9596085829811408),
    ('horizon', 1800, 0.8329070417832541),
    ('ground', 180, 0.719823559106309)
])
def test_marion_integrate_scalar(region, N, expected):
    actual = _iam.marion_integrate(_iam.ashrae, 20, region, N)
    assert_allclose(actual, expected)

    with np.errstate(invalid='ignore'):
        actual = _iam.marion_integrate(_iam.ashrae, np.nan, region, N)
    expected = np.nan
    assert_allclose(actual, expected)


@pytest.mark.parametrize('region,N,expected', [
    ('sky', 180, [0.9523611991069362, 0.9596085829811408, 0.9619811198105501]),
    ('horizon', 1800, [0.0, 0.8329070417832541, 0.8987287652347437]),
    ('ground', 180, [0.0, 0.719823559106309, 0.8186360238536674])
])
def test_marion_integrate_list(region, N, expected):
    actual = _iam.marion_integrate(_iam.ashrae, [0, 20, 30], region, N)
    assert_allclose(actual, expected)

    with np.errstate(invalid='ignore'):
        actual = _iam.marion_integrate(_iam.ashrae, [0, 20, np.nan], region, N)
    assert_allclose(actual, [expected[0], expected[1], np.nan])


@pytest.mark.parametrize('region,N,expected', [
    ('sky', 180, [0.9523611991069362, 0.9596085829811408, 0.9619811198105501]),
    ('horizon', 1800, [0.0, 0.8329070417832541, 0.8987287652347437]),
    ('ground', 180, [0.0, 0.719823559106309, 0.8186360238536674])
])
def test_marion_integrate_series(region, N, expected):
    idx = pd.date_range('2019-01-01', periods=3, freq='h')
    tilt = pd.Series([0, 20, 30], index=idx)
    expected = pd.Series(expected, index=idx)
    actual = _iam.marion_integrate(_iam.ashrae, tilt, region, N)
    assert_series_equal(actual, expected)

    tilt.iloc[1] = np.nan
    expected.iloc[1] = np.nan
    with np.errstate(invalid='ignore'):
        actual = _iam.marion_integrate(_iam.ashrae, tilt, region, N)
    assert_allclose(actual, expected)


def test_marion_integrate_ground_flat():
    iam = _iam.marion_integrate(_iam.ashrae, 0, 'horizon', 1800)
    assert_allclose(iam, 0)


def test_marion_integrate_invalid():
    # check for invalid region string.  this actually gets checked twice,
    # with the difference being whether `num` is specified or not.
    with pytest.raises(ValueError):
        _iam.marion_integrate(_iam.ashrae, 0, 'bad')

    with pytest.raises(ValueError):
        _iam.marion_integrate(_iam.ashrae, 0, 'bad', 180)


def test_schlick():
    idx = pd.date_range('2019-01-01', freq='h', periods=9)
    aoi = pd.Series([-180, -135, -90, -45, 0, 45, 90, 135, 180], idx)
    expected = pd.Series([0, 0, 0, 0.99784451, 1.0, 0.99784451, 0, 0, 0], idx)

    # scalars
    for aoi_scalar, expected_scalar in zip(aoi, expected):
        actual = _iam.schlick(aoi_scalar)
        assert_allclose(expected_scalar, actual)

    # numpy arrays
    actual = _iam.schlick(aoi.values)
    assert_allclose(expected.values, actual)

    # pandas Series
    actual = _iam.schlick(aoi)
    assert_series_equal(expected, actual)


def test_schlick_diffuse():
    surface_tilt = np.array([0, 20, 70, 90])
    # expected values calculated with marion_integrate and schlick
    expected_sky = np.array([0.95238092, 0.96249934, 0.96228167, 0.95238094])
    expected_ground = np.array([0, 0.62693858, 0.93218737, 0.95238094])

    # numpy arrays
    actual_sky, actual_ground = _iam.schlick_diffuse(surface_tilt)
    assert_allclose(expected_sky, actual_sky)
    assert_allclose(expected_ground, actual_ground, rtol=1e-6)

    # scalars
    for i in range(len(surface_tilt)):
        actual_sky, actual_ground = _iam.schlick_diffuse(surface_tilt[i])
        assert_allclose(expected_sky[i], actual_sky)
        assert_allclose(expected_ground[i], actual_ground, rtol=1e-6)

    # pandas Series
    idx = pd.date_range('2019-01-01', freq='h', periods=len(surface_tilt))
    actual_sky, actual_ground = _iam.schlick_diffuse(pd.Series(surface_tilt,
                                                               idx))
    assert_series_equal(pd.Series(expected_sky, idx), actual_sky)
    assert_series_equal(pd.Series(expected_ground, idx), actual_ground,
                        rtol=1e-6)
